@@ -14,27 +14,32 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-import functools
+import datetime
+import typing as ty
+
+import requests
 
 from keystoneauth1 import _utils as utils
 from keystoneauth1.access import service_catalog
 from keystoneauth1.access import service_providers
-
+from keystoneauth1.access import types
 
 # gap, in seconds, to determine whether the given token is about to expire
 STALE_TOKEN_DURATION = 30
 
-
-__all__ = ('AccessInfo',
-           'AccessInfoV2',
-           'AccessInfoV3',
-           'create')
+__all__ = ('AccessInfo', 'AccessInfoV2', 'AccessInfoV3', 'create')
 
 
-def create(resp=None, body=None, auth_token=None):
+def create(
+    resp: ty.Optional[requests.Response] = None,
+    body: ty.Optional[ty.Dict[str, object]] = None,
+    auth_token: ty.Optional[str] = None,
+) -> 'AccessInfo':
     if resp and not body:
         body = resp.json()
 
+    assert body is not None  # nosec B101
+
     if 'token' in body:
         if resp and not auth_token:
             auth_token = resp.headers.get('X-Subject-Token')
@@ -46,55 +51,54 @@ def create(resp=None, body=None, auth_token=None):
     raise ValueError('Unrecognized auth response')
 
 
-def _missingproperty(f):
-
-    @functools.wraps(f)
-    def inner(self):
-        try:
-            return f(self)
-        except KeyError:
-            return None
-
-    return property(inner)
-
-
-class AccessInfo(object):
+class AccessInfo:
     """Encapsulates a raw authentication token from keystone.
 
     Provides helper methods for extracting useful values from that token.
-
     """
 
-    _service_catalog_class = None
+    _service_catalog_class: ty.Type[service_catalog.ServiceCatalog]
+    _data: ty.Any
 
-    def __init__(self, body, auth_token=None):
+    def __init__(
+        self, body: ty.Dict[str, ty.Any], auth_token: ty.Optional[str] = None
+    ):
         self._data = body
         self._auth_token = auth_token
-        self._service_catalog = None
-        self._service_providers = None
+        self._service_catalog: ty.Optional[service_catalog.ServiceCatalog] = (
+            None
+        )
+        self._service_providers: ty.Optional[
+            service_providers.ServiceProviders
+        ] = None
 
     @property
-    def service_catalog(self):
+    def service_catalog(self) -> service_catalog.ServiceCatalog:
         if not self._service_catalog:
             self._service_catalog = self._service_catalog_class.from_token(
-                self._data)
+                self._data
+            )
 
         return self._service_catalog
 
-    def will_expire_soon(self, stale_duration=STALE_TOKEN_DURATION):
+    def will_expire_soon(
+        self, stale_duration: int = STALE_TOKEN_DURATION
+    ) -> bool:
         """Determine if expiration is about to occur.
 
         :returns: true if expiration is within the given duration
         :rtype: boolean
-
         """
+        if self.expires is None:
+            return False
+
         norm_expires = utils.normalize_time(self.expires)
         # (gyee) should we move auth_token.will_expire_soon() to timeutils
         # instead of duplicating code here?
         soon = utils.from_utcnow(seconds=stale_duration)
         return norm_expires < soon
 
-    def has_service_catalog(self):
+    def has_service_catalog(self) -> bool:
         """Return true if the auth token has a service catalog.
 
         :returns: boolean
@@ -102,7 +106,7 @@ class AccessInfo(object):
         raise NotImplementedError()
 
     @property
-    def auth_token(self):
+    def auth_token(self) -> ty.Optional[str]:
         """Return the token_id associated with the auth request.
 
         To be used in headers for authenticating OpenStack API requests.
@@ -112,7 +116,7 @@ class AccessInfo(object):
         return self._auth_token
 
     @property
-    def expires(self):
+    def expires(self) -> ty.Optional[datetime.datetime]:
         """Return the token expiration (as datetime object).
 
         :returns: datetime
@@ -120,7 +124,7 @@ class AccessInfo(object):
         raise NotImplementedError()
 
     @property
-    def issued(self):
+    def issued(self) -> ty.Optional[datetime.datetime]:
         """Return the token issue time (as datetime object).
 
         :returns: datetime
@@ -128,7 +132,7 @@ class AccessInfo(object):
         raise NotImplementedError()
 
     @property
-    def username(self):
+    def username(self) -> ty.Optional[str]:
         """Return the username associated with the auth request.
 
         Follows the pattern defined in the V2 API of first looking for 'name',
@@ -140,7 +144,7 @@ class AccessInfo(object):
         raise NotImplementedError()
 
     @property
-    def user_id(self):
+    def user_id(self) -> ty.Optional[str]:
         """Return the user id associated with the auth request.
 
         :returns: str
@@ -148,7 +152,7 @@ class AccessInfo(object):
         raise NotImplementedError()
 
     @property
-    def user_domain_id(self):
+    def user_domain_id(self) -> ty.Optional[str]:
         """Return the user's domain id associated with the auth request.
 
         :returns: str
@@ -156,7 +160,7 @@ class AccessInfo(object):
         raise NotImplementedError()
 
     @property
-    def user_domain_name(self):
+    def user_domain_name(self) -> ty.Optional[str]:
         """Return the user's domain name associated with the auth request.
 
         :returns: str
@@ -164,7 +168,7 @@ class AccessInfo(object):
         raise NotImplementedError()
 
     @property
-    def role_ids(self):
+    def role_ids(self) -> ty.Optional[ty.List[str]]:
         """Return a list of user's role ids associated with the auth request.
 
         :returns: a list of strings of role ids
@@ -172,7 +176,7 @@ class AccessInfo(object):
         raise NotImplementedError()
 
     @property
-    def role_names(self):
+    def role_names(self) -> ty.Optional[ty.List[str]]:
         """Return a list of user's role names associated with the auth request.
 
         :returns: a list of strings of role names
@@ -180,7 +184,7 @@ class AccessInfo(object):
         raise NotImplementedError()
 
     @property
-    def domain_name(self):
+    def domain_name(self) -> ty.Optional[str]:
         """Return the domain name associated with the auth request.
 
         :returns: str or None (if no domain associated with the token)
@@ -188,7 +192,7 @@ class AccessInfo(object):
         raise NotImplementedError()
 
     @property
-    def domain_id(self):
+    def domain_id(self) -> ty.Optional[str]:
         """Return the domain id associated with the auth request.
 
         :returns: str or None (if no domain associated with the token)
@@ -196,7 +200,7 @@ class AccessInfo(object):
         raise NotImplementedError()
 
     @property
-    def project_name(self):
+    def project_name(self) -> ty.Optional[str]:
         """Return the project name associated with the auth request.
 
         :returns: str or None (if no project associated with the token)
@@ -204,12 +208,12 @@ class AccessInfo(object):
         raise NotImplementedError()
 
     @property
-    def tenant_name(self):
+    def tenant_name(self) -> ty.Optional[str]:
         """Synonym for project_name."""
         return self.project_name
 
     @property
-    def scoped(self):
+    def scoped(self) -> bool:
         """Return true if the auth token was scoped.
 
         Returns true if scoped to a tenant(project) or domain,
@@ -222,7 +226,7 @@ class AccessInfo(object):
         return self.project_scoped or self.domain_scoped or self.system_scoped
 
     @property
-    def project_scoped(self):
+    def project_scoped(self) -> bool:
         """Return true if the auth token was scoped to a tenant (project).
 
         :returns: bool
@@ -230,7 +234,7 @@ class AccessInfo(object):
         return bool(self.project_id)
 
     @property
-    def domain_scoped(self):
+    def domain_scoped(self) -> bool:
         """Return true if the auth token was scoped to a domain.
 
         :returns: bool
@@ -238,7 +242,7 @@ class AccessInfo(object):
         raise NotImplementedError()
 
     @property
-    def system_scoped(self):
+    def system_scoped(self) -> bool:
         """Return true if the auth token was scoped to the system.
 
         :returns: bool
@@ -246,7 +250,7 @@ class AccessInfo(object):
         raise NotImplementedError()
 
     @property
-    def trust_id(self):
+    def trust_id(self) -> ty.Optional[str]:
         """Return the trust id associated with the auth request.
 
         :returns: str or None (if no trust associated with the token)
@@ -254,7 +258,7 @@ class AccessInfo(object):
         raise NotImplementedError()
 
     @property
-    def trust_scoped(self):
+    def trust_scoped(self) -> bool:
         """Return true if the auth token was scoped from a delegated trust.
 
         The trust delegation is via the OS-TRUST v3 extension.
@@ -264,7 +268,7 @@ class AccessInfo(object):
         raise NotImplementedError()
 
     @property
-    def trustee_user_id(self):
+    def trustee_user_id(self) -> ty.Optional[str]:
         """Return the trustee user id associated with a trust.
 
         :returns: str or None (if no trust associated with the token)
@@ -272,7 +276,7 @@ class AccessInfo(object):
         raise NotImplementedError()
 
     @property
-    def trustor_user_id(self):
+    def trustor_user_id(self) -> ty.Optional[str]:
         """Return the trustor user id associated with a trust.
 
         :returns: str or None (if no trust associated with the token)
@@ -280,7 +284,7 @@ class AccessInfo(object):
         raise NotImplementedError()
 
     @property
-    def project_id(self):
+    def project_id(self) -> ty.Optional[str]:
         """Return the project ID associated with the auth request.
 
         This returns None if the auth token wasn't scoped to a project.
@@ -290,12 +294,12 @@ class AccessInfo(object):
         raise NotImplementedError()
 
     @property
-    def tenant_id(self):
+    def tenant_id(self) -> ty.Optional[str]:
         """Synonym for project_id."""
         return self.project_id
 
     @property
-    def project_domain_id(self):
+    def project_domain_id(self) -> ty.Optional[str]:
         """Return the project's domain id associated with the auth request.
 
         :returns: str
@@ -303,7 +307,7 @@ class AccessInfo(object):
         raise NotImplementedError()
 
     @property
-    def project_domain_name(self):
+    def project_domain_name(self) -> ty.Optional[str]:
         """Return the project's domain name associated with the auth request.
 
         :returns: str
@@ -311,7 +315,7 @@ class AccessInfo(object):
         raise NotImplementedError()
 
     @property
-    def oauth_access_token_id(self):
+    def oauth_access_token_id(self) -> ty.Optional[str]:
         """Return the access token ID if OAuth authentication used.
 
         :returns: str or None.
@@ -319,7 +323,7 @@ class AccessInfo(object):
         raise NotImplementedError()
 
     @property
-    def oauth_consumer_id(self):
+    def oauth_consumer_id(self) -> ty.Optional[str]:
         """Return the consumer ID if OAuth authentication used.
 
         :returns: str or None.
@@ -327,7 +331,7 @@ class AccessInfo(object):
         raise NotImplementedError()
 
     @property
-    def is_federated(self):
+    def is_federated(self) -> bool:
         """Return true if federation was used to get the token.
 
         :returns: boolean
@@ -335,7 +339,7 @@ class AccessInfo(object):
         raise NotImplementedError()
 
     @property
-    def is_admin_project(self):
+    def is_admin_project(self) -> bool:
         """Return true if the current project scope is the admin project.
 
         For backwards compatibility purposes if there is nothing specified in
@@ -347,7 +351,7 @@ class AccessInfo(object):
         raise NotImplementedError()
 
     @property
-    def audit_id(self):
+    def audit_id(self) -> ty.Optional[str]:
         """Return the audit ID if present.
 
         :returns: str or None.
@@ -355,7 +359,7 @@ class AccessInfo(object):
         raise NotImplementedError()
 
     @property
-    def audit_chain_id(self):
+    def audit_chain_id(self) -> ty.Optional[str]:
         """Return the audit chain ID if present.
 
         In the event that a token was rescoped then this ID will be the
@@ -367,7 +371,7 @@ class AccessInfo(object):
         raise NotImplementedError()
 
     @property
-    def initial_audit_id(self):
+    def initial_audit_id(self) -> ty.Optional[str]:
         """The audit ID of the initially requested token.
 
         This is the :py:attr:`audit_chain_id` if present or the
@@ -376,7 +380,9 @@ class AccessInfo(object):
         return self.audit_chain_id or self.audit_id
 
     @property
-    def service_providers(self):
+    def service_providers(
+        self,
+    ) -> ty.Optional[service_providers.ServiceProviders]:
         """Return an object representing the list of trusted service providers.
 
         Used for Keystone2Keystone federating-out.
@@ -387,7 +393,7 @@ class AccessInfo(object):
         raise NotImplementedError()
 
     @property
-    def bind(self):
+    def bind(self) -> ty.Optional[ty.Dict[str, ty.Any]]:
         """Information about external mechanisms the token is bound to.
 
         If a token is bound to an external authentication mechanism it can only
@@ -403,7 +409,7 @@ class AccessInfo(object):
         raise NotImplementedError()
 
     @property
-    def project_is_domain(self):
+    def project_is_domain(self) -> ty.Optional[bool]:
         """Return if a project act as a domain.
 
         :returns: bool
@@ -416,184 +422,175 @@ class AccessInfoV2(AccessInfo):
 
     version = 'v2.0'
     _service_catalog_class = service_catalog.ServiceCatalogV2
+    _data: types.TokenResponseV2
 
-    def has_service_catalog(self):
-        return 'serviceCatalog' in self._data.get('access', {})
-
-    @_missingproperty
-    def auth_token(self):
-        set_token = super(AccessInfoV2, self).auth_token
-        return set_token or self._data['access']['token']['id']
+    def has_service_catalog(self) -> bool:
+        return 'serviceCatalog' in self._data['access']
 
     @property
-    def _token(self):
+    def _token(self) -> types.TokenV2:
         return self._data['access']['token']
 
-    @_missingproperty
-    def expires(self):
-        return utils.parse_isotime(self._token.get('expires'))
+    @property
+    def auth_token(self) -> ty.Optional[str]:
+        set_token = super().auth_token
+        return set_token or self._token.get('id')
 
-    @_missingproperty
-    def issued(self):
+    @property
+    def expires(self) -> ty.Optional[datetime.datetime]:
+        return utils.parse_isotime(self._token['expires'])
+
+    @property
+    def issued(self) -> ty.Optional[datetime.datetime]:
         return utils.parse_isotime(self._token['issued_at'])
 
     @property
-    def _user(self):
+    def _user(self) -> types.UserV2:
         return self._data['access']['user']
 
-    @_missingproperty
-    def username(self):
+    @property
+    def username(self) -> ty.Optional[str]:
         return self._user.get('name') or self._user.get('username')
 
-    @_missingproperty
-    def user_id(self):
-        return self._user['id']
+    @property
+    def user_id(self) -> ty.Optional[str]:
+        return self._user.get('id')
 
     @property
-    def user_domain_id(self):
+    def user_domain_id(self) -> ty.Optional[str]:
         return None
 
     @property
-    def user_domain_name(self):
+    def user_domain_name(self) -> ty.Optional[str]:
         return None
 
-    @_missingproperty
-    def role_ids(self):
-        metadata = self._data.get('access', {}).get('metadata', {})
+    @property
+    def role_ids(self) -> ty.Optional[ty.List[str]]:
+        metadata = self._data['access'].get('metadata', {})
         return metadata.get('roles', [])
 
-    @_missingproperty
-    def role_names(self):
+    @property
+    def role_names(self) -> ty.Optional[ty.List[str]]:
         return [r['name'] for r in self._user.get('roles', [])]
 
     @property
-    def domain_name(self):
+    def domain_name(self) -> ty.Optional[str]:
         return None
 
     @property
-    def domain_id(self):
+    def domain_id(self) -> ty.Optional[str]:
         return None
 
     @property
-    def project_name(self):
-        try:
-            tenant_dict = self._token['tenant']
-        except KeyError:
-            pass
-        else:
-            return tenant_dict.get('name')
+    def project_name(self) -> ty.Optional[str]:
+        if 'tenant' in self._token:
+            return self._token['tenant'].get('name')
 
         # pre grizzly
-        try:
+        if 'tenantName' in self._user:
             return self._user['tenantName']
-        except KeyError:
-            pass
 
         # pre diablo, keystone only provided a tenantId
-        try:
-            return self._token['tenantId']
-        except KeyError:
-            pass
+        if 'tenantId' in self._token:
+            return self._token['tenantId']  # type: ignore
+
+        return None
 
     @property
-    def domain_scoped(self):
+    def domain_scoped(self) -> bool:
         return False
 
     @property
-    def system_scoped(self):
+    def system_scoped(self) -> bool:
         return False
 
     @property
-    def _trust(self):
-        return self._data['access']['trust']
+    def _trust(self) -> ty.Optional[types.TrustV2]:
+        return self._data['access'].get('trust')
 
-    @_missingproperty
-    def trust_id(self):
-        return self._trust['id']
+    @property
+    def trust_id(self) -> ty.Optional[str]:
+        return self._trust and self._trust['id']
 
-    @_missingproperty
-    def trust_scoped(self):
+    @property
+    def trust_scoped(self) -> bool:
         return bool(self._trust)
 
-    @_missingproperty
-    def trustee_user_id(self):
-        return self._trust['trustee_user_id']
+    @property
+    def trustee_user_id(self) -> ty.Optional[str]:
+        return self._trust and self._trust['trustee_user_id']
 
     @property
-    def trustor_user_id(self):
+    def trustor_user_id(self) -> ty.Optional[str]:
         # this information is not available in the v2 token bug: #1331882
         return None
 
     @property
-    def project_id(self):
-        try:
-            tenant_dict = self._token['tenant']
-        except KeyError:
-            pass
-        else:
-            return tenant_dict.get('id')
+    def project_id(self) -> ty.Optional[str]:
+        if 'tenant' in self._token:
+            return self._token['tenant'].get('id')
 
         # pre grizzly
-        try:
+        if 'tenantId' in self._user:
             return self._user['tenantId']
-        except KeyError:
-            pass
 
-        # pre diablo
-        try:
-            return self._token['tenantId']
-        except KeyError:
-            pass
+        # pre diablo, keystone only provided a tenantId
+        if 'tenantId' in self._token:
+            return self._token['tenantId']  # type: ignore
+
+        return None
 
     @property
-    def project_is_domain(self):
+    def project_is_domain(self) -> ty.Optional[bool]:
         return False
 
     @property
-    def project_domain_id(self):
+    def project_domain_id(self) -> ty.Optional[str]:
         return None
 
     @property
-    def project_domain_name(self):
+    def project_domain_name(self) -> ty.Optional[str]:
         return None
 
     @property
-    def oauth_access_token_id(self):
+    def oauth_access_token_id(self) -> ty.Optional[str]:
         return None
 
     @property
-    def oauth_consumer_id(self):
+    def oauth_consumer_id(self) -> ty.Optional[str]:
         return None
 
     @property
-    def is_federated(self):
+    def is_federated(self) -> bool:
         return False
 
     @property
-    def is_admin_project(self):
+    def is_admin_project(self) -> bool:
         return True
 
     @property
-    def audit_id(self):
+    def audit_id(self) -> ty.Optional[str]:
         try:
             return self._token.get('audit_ids', [])[0]
         except IndexError:
             return None
 
     @property
-    def audit_chain_id(self):
+    def audit_chain_id(self) -> ty.Optional[str]:
         try:
             return self._token.get('audit_ids', [])[1]
         except IndexError:
             return None
 
     @property
-    def service_providers(self):
+    def service_providers(
+        self,
+    ) -> ty.Optional[service_providers.ServiceProviders]:
         return None
 
-    @_missingproperty
-    def bind(self):
-        return self._token['bind']
+    @property
+    def bind(self) -> ty.Optional[ty.Dict[str, ty.Any]]:
+        return self._token.get('bind')
 
 
 class AccessInfoV3(AccessInfo):
@@ -601,172 +598,217 @@ class AccessInfoV3(AccessInfo):
 
     version = 'v3'
     _service_catalog_class = service_catalog.ServiceCatalogV3
-
-    def has_service_catalog(self):
-        return 'catalog' in self._data['token']
+    _data: types.TokenResponseV3
 
     @property
-    def _user(self):
-        return self._data['token']['user']
+    def _token(self) -> types.TokenV3:
+        return self._data['token']
+
+    def has_service_catalog(self) -> bool:
+        return 'catalog' in self._token
 
     @property
-    def is_federated(self):
-        return 'OS-FEDERATION' in self._user
+    def expires(self) -> ty.Optional[datetime.datetime]:
+        return utils.parse_isotime(self._token['expires_at'])
 
     @property
-    def is_admin_project(self):
-        return self._data.get('token', {}).get('is_admin_project', True)
+    def issued(self) -> ty.Optional[datetime.datetime]:
+        return utils.parse_isotime(self._token['issued_at'])
 
-    @_missingproperty
-    def expires(self):
-        return utils.parse_isotime(self._data['token']['expires_at'])
+    @property
+    def _user(self) -> types.UserV3:
+        return self._token['user']
 
-    @_missingproperty
-    def issued(self):
-        return utils.parse_isotime(self._data['token']['issued_at'])
+    @property
+    def username(self) -> ty.Optional[str]:
+        return self._user['name']
 
-    @_missingproperty
-    def user_id(self):
+    @property
+    def user_id(self) -> ty.Optional[str]:
         return self._user['id']
 
     @property
-    def user_domain_id(self):
-        try:
-            return self._user['domain']['id']
-        except KeyError:
-            if self.is_federated:
-                return None
-            raise
+    def _user_domain(self) -> types.UserDomainV3:
+        return self._user['domain']
 
     @property
-    def user_domain_name(self):
-        try:
-            return self._user['domain']['name']
-        except KeyError:
-            if self.is_federated:
-                return None
-            raise
-
-    @_missingproperty
-    def role_ids(self):
-        return [r['id'] for r in self._data['token'].get('roles', [])]
-
-    @_missingproperty
-    def role_names(self):
-        return [r['name'] for r in self._data['token'].get('roles', [])]
-
-    @_missingproperty
-    def username(self):
-        return self._user['name']
-
-    @_missingproperty
-    def system(self):
-        return self._data['token']['system']
+    def user_domain_id(self) -> ty.Optional[str]:
+        return self._user['domain']['id']
 
     @property
-    def _domain(self):
-        return self._data['token']['domain']
-
-    @_missingproperty
-    def domain_name(self):
-        return self._domain['name']
-
-    @_missingproperty
-    def domain_id(self):
-        return self._domain['id']
+    def user_domain_name(self) -> ty.Optional[str]:
+        return self._user['domain']['name']
 
     @property
-    def _project(self):
-        return self._data['token']['project']
-
-    @_missingproperty
-    def project_id(self):
-        return self._project['id']
-
-    @_missingproperty
-    def project_is_domain(self):
-        return self._data['token']['is_domain']
-
-    @_missingproperty
-    def project_domain_id(self):
-        return self._project['domain']['id']
-
-    @_missingproperty
-    def project_domain_name(self):
-        return self._project['domain']['name']
-
-    @_missingproperty
-    def project_name(self):
-        return self._project['name']
+    def role_ids(self) -> ty.Optional[ty.List[str]]:
+        return [r['id'] for r in self._token.get('roles', [])]
 
     @property
-    def domain_scoped(self):
-        try:
-            return bool(self._domain)
-        except KeyError:
-            return False
-
-    @_missingproperty
-    def system_scoped(self):
-        return self._data['token']['system'].get('all', False)
+    def role_names(self) -> ty.Optional[ty.List[str]]:
+        return [r['name'] for r in self._token.get('roles', [])]
 
     @property
-    def _trust(self):
-        return self._data['token']['OS-TRUST:trust']
-
-    @_missingproperty
-    def trust_id(self):
-        return self._trust['id']
+    def system(self) -> ty.Optional[types.SystemV3]:
+        return self._token.get('system')
 
     @property
-    def trust_scoped(self):
-        try:
-            return bool(self._trust)
-        except KeyError:
-            return False
-
-    @_missingproperty
-    def trustee_user_id(self):
-        return self._trust['trustee_user']['id']
-
-    @_missingproperty
-    def trustor_user_id(self):
-        return self._trust['trustor_user']['id']
+    def _domain(self) -> ty.Optional[types.DomainV3]:
+        # only present for domain-scoped tokens
+        return self._token.get('domain')
 
     @property
-    def _oauth(self):
-        return self._data['token']['OS-OAUTH1']
-
-    @_missingproperty
-    def oauth_access_token_id(self):
-        return self._oauth['access_token_id']
-
-    @_missingproperty
-    def oauth_consumer_id(self):
-        return self._oauth['consumer_id']
-
-    @_missingproperty
-    def audit_id(self):
-        try:
-            return self._data['token']['audit_ids'][0]
-        except IndexError:
-            return None
-
-    @_missingproperty
-    def audit_chain_id(self):
-        try:
-            return self._data['token']['audit_ids'][1]
-        except IndexError:
-            return None
+    def domain_name(self) -> ty.Optional[str]:
+        return self._domain and self._domain['name']
 
     @property
-    def service_providers(self):
+    def domain_id(self) -> ty.Optional[str]:
+        return self._domain and self._domain['id']
+
+    @property
+    def _project(self) -> ty.Optional[types.ProjectV3]:
+        # only present for project-scoped tokens
+        return self._token.get('project')
+
+    @property
+    def project_id(self) -> ty.Optional[str]:
+        return self._project and self._project.get('id')
+
+    @property
+    def project_name(self) -> ty.Optional[str]:
+        return self._project and self._project.get('name')
+
+    @property
+    def project_is_domain(self) -> ty.Optional[bool]:
+        return self._token.get('is_domain')
+
+    @property
+    def _project_domain(self) -> ty.Optional[types.ProjectDomainV3]:
+        return self._project and self._project.get('domain')
+
+    @property
+    def project_domain_id(self) -> ty.Optional[str]:
+        return self._project_domain and self._project_domain['id']
+
+    @property
+    def project_domain_name(self) -> ty.Optional[str]:
+        return self._project_domain and self._project_domain['name']
+
+    @property
+    def domain_scoped(self) -> bool:
+        return bool(self._domain)
+
+    @property
+    def system_scoped(self) -> bool:
+        return bool(self._token.get('system', {}).get('all'))
+
+    @property
+    def _trust(self) -> ty.Optional[types.TrustV3]:
+        # only present for trust-scoped tokens
+        return self._token.get('OS-TRUST:trust')
+
+    @property
+    def trust_id(self) -> ty.Optional[str]:
+        return self._trust and self._trust['id']
+
+    @property
+    def trust_scoped(self) -> bool:
+        return bool(self._trust)
+
+    @property
+    def trustee_user_id(self) -> ty.Optional[str]:
+        return self._trust and self._trust['trustee_user']['id']
+
+    @property
+    def trustor_user_id(self) -> ty.Optional[str]:
+        return self._trust and self._trust['trustor_user']['id']
+
+    # TODO(stephenfin): Should this be private like every other high-level
+    # accessor? As this stands, it can raise KeyError
+    @property
+    def application_credential(self) -> types.ApplicationCredentialV3:
+        return self._token['application_credential']
+
+    @property
+    def _application_credential(
+        self,
+    ) -> ty.Optional[types.ApplicationCredentialV3]:
+        # only present if user has authenticated with application credentials
+        return self._token.get('application_credential')
+
+    @property
+    def application_credential_id(self) -> ty.Optional[str]:
+        return (
+            self._application_credential
+            and self._application_credential.get('id')
+        )
+
+    @property
+    def application_credential_access_rules(
+        self,
+    ) -> ty.Optional[ty.List[types.ApplicationCredentialAccessRuleV3]]:
+        return (
+            self._application_credential
+            and self._application_credential.get('access_rules')
+        )
+
+    @property
+    def _oauth(self) -> ty.Optional[types.OAuth1V3]:
+        # only present if user has authenticated with OAuth1
+        return self._token.get('OS-OAUTH1')
+
+    @property
+    def oauth_access_token_id(self) -> ty.Optional[str]:
+        return self._oauth and self._oauth.get('access_token_id')
+
+    @property
+    def oauth_consumer_id(self) -> ty.Optional[str]:
+        return self._oauth and self._oauth.get('consumer_id')
+
+    @property
+    def is_federated(self) -> bool:
+        return 'OS-FEDERATION' in self._user
+
+    @property
+    def is_admin_project(self) -> bool:
+        return bool(self._token.get('is_admin_project', True))
+
+    @property
+    def audit_id(self) -> ty.Optional[str]:
+        ret = self._token.get('audit_ids', [])
+        return ret[0] if ret else None
+
+    @property
+    def audit_chain_id(self) -> ty.Optional[str]:
+        ret = self._token.get('audit_ids', [])
+        return ret[1] if len(ret) > 1 else None
+
+    @property
+    def service_providers(
+        self,
+    ) -> ty.Optional[service_providers.ServiceProviders]:
         if not self._service_providers:
             self._service_providers = (
-                service_providers.ServiceProviders.from_token(self._data))
+                service_providers.ServiceProviders.from_token(self._data)
+            )
 
         return self._service_providers
 
-    @_missingproperty
-    def bind(self):
-        return self._data['token']['bind']
+    @property
+    def bind(self) -> ty.Optional[ty.Dict[str, ty.Any]]:
+        return self._token.get('bind')
+
+    # TODO(stephenfin): Should this be private like every other high-level
+    # accessor? As this stands, it can raise KeyError
+    @property
+    def oauth2_credential(self) -> types.OAuth2V3:
+        return self._token['oauth2_credential']
+
+    @property
+    def _oauth2_credential(self) -> ty.Optional[types.OAuth2V3]:
+        return self._token.get('oauth2_credential')
+
+    @property
+    def oauth2_credential_thumbprint(self) -> ty.Optional[str]:
+        return self._oauth2_credential and self._oauth2_credential.get(
+            'x5t#S256'
+        )
