@@ -18,22 +18,38 @@
 
 import abc
 import copy
+import typing as ty
+
+import typing_extensions as ty_ext
 
 from keystoneauth1 import discover
 from keystoneauth1 import exceptions
 
+_SERVICE_CATALOG_T = ty.List[ty.Dict[str, ty.Any]]
+
 
 class ServiceCatalog(metaclass=abc.ABCMeta):
     """Helper methods for dealing with a Keystone Service Catalog."""
 
-    def __init__(self, catalog):
+    def __init__(self, catalog: _SERVICE_CATALOG_T):
         self._catalog = catalog
 
-    def _get_endpoint_region(self, endpoint):
+    @classmethod
+    @abc.abstractmethod
+    def from_token(cls, token: ty.Dict[str, ty.Any]) -> ty_ext.Self:
+        """Retrieve the service catalog from a token.
+
+        :param token:
+        :returns: A service catalog.
+        """
+
+    def _get_endpoint_region(
+        self, endpoint: ty.Dict[str, str]
+    ) -> ty.Optional[str]:
         return endpoint.get('region_id') or endpoint.get('region')
 
     @property
-    def catalog(self):
+    def catalog(self) -> _SERVICE_CATALOG_T:
         """Return the raw service catalog content, mostly useful for debugging.
 
         Applications should avoid this and use accessor methods instead.
@@ -43,7 +59,9 @@ class ServiceCatalog(metaclass=abc.ABCMeta):
         return self._catalog
 
     @abc.abstractmethod
-    def is_interface_match(self, endpoint, interface):
+    def is_interface_match(
+        self, endpoint: ty.Dict[str, str], interface: str
+    ) -> bool:
         """Helper function to normalize endpoint matching across v2 and v3.
 
         :returns: True if the provided endpoint matches the required
@@ -51,7 +69,7 @@ class ServiceCatalog(metaclass=abc.ABCMeta):
         """
 
     @staticmethod
-    def normalize_interface(self, interface):
+    def normalize_interface(interface: str) -> str:
         """Handle differences in the way v2 and v3 catalogs specify endpoint.
 
         Both v2 and v3 must be able to handle the endpoint style of the other.
@@ -63,7 +81,9 @@ class ServiceCatalog(metaclass=abc.ABCMeta):
         """
         return interface
 
-    def _normalize_endpoints(self, endpoints):
+    def _normalize_endpoints(
+        self, endpoints: ty.List[ty.Dict[str, ty.Any]]
+    ) -> ty.List[ty.Dict[str, ty.Any]]:
         """Translate endpoint description dicts into v3 form.
 
         Takes a raw endpoint description from the catalog and changes
@@ -83,7 +103,9 @@ class ServiceCatalog(metaclass=abc.ABCMeta):
             new_endpoints.append(new_endpoint)
         return new_endpoints
 
-    def _denormalize_endpoints(self, endpoints):
+    def _denormalize_endpoints(
+        self, endpoints: ty.List[discover.EndpointData]
+    ) -> ty.List[ty.Union[str, None]]:
         """Return original endpoint description dicts.
 
         Takes a list of EndpointData objects and returns the original
@@ -95,7 +117,7 @@ class ServiceCatalog(metaclass=abc.ABCMeta):
         """
         return [endpoint.raw_endpoint for endpoint in endpoints]
 
-    def normalize_catalog(self):
+    def normalize_catalog(self) -> ty.List[ty.Dict[str, ty.Any]]:
         """Return the catalog normalized into v3 format."""
         catalog = []
         for service in copy.deepcopy(self._catalog):
@@ -114,7 +136,8 @@ class ServiceCatalog(metaclass=abc.ABCMeta):
             service.setdefault('id', None)
 
             service['endpoints'] = self._normalize_endpoints(
-                service.get('endpoints', []))
+                service.get('endpoints', [])
+            )
 
             for endpoint in service['endpoints']:
                 endpoint['region_name'] = self._get_endpoint_region(endpoint)
@@ -122,16 +145,29 @@ class ServiceCatalog(metaclass=abc.ABCMeta):
             catalog.append(service)
         return catalog
 
-    def _get_interface_list(self, interface):
+    def _get_interface_list(
+        self,
+        interface: ty.Union[
+            str, ty.List[str], ty.Tuple[str], ty.Set[str], None
+        ],
+    ) -> ty.List[str]:
         if not interface:
             return []
         if not isinstance(interface, (list, tuple, set)):
             interface = [interface]
         return [self.normalize_interface(i) for i in interface]
 
-    def get_endpoints_data(self, service_type=None, interface=None,
-                           region_name=None, service_name=None,
-                           service_id=None, endpoint_id=None):
+    def get_endpoints_data(
+        self,
+        service_type: ty.Optional[str] = None,
+        interface: ty.Union[
+            str, ty.List[str], ty.Tuple[str], ty.Set[str], None
+        ] = None,
+        region_name: ty.Optional[str] = None,
+        service_name: ty.Optional[str] = None,
+        service_id: ty.Optional[str] = None,
+        endpoint_id: ty.Optional[str] = None,
+    ) -> ty.Dict[str, ty.List[discover.EndpointData]]:
         """Fetch and filter endpoint data for the specified service(s).
 
         Returns endpoints for the specified service (or all) containing
@@ -154,27 +190,26 @@ class ServiceCatalog(metaclass=abc.ABCMeta):
         :param string service_id: The identifier of a service.
         :param string endpoint_id: The identifier of an endpoint.
 
-        :returns: a list of matching EndpointData objects
-        :rtype: list(`keystoneauth1.discover.EndpointData`)
-
         :returns: a dict, keyed by service_type, of lists of EndpointData
         """
         interfaces = self._get_interface_list(interface)
 
-        matching_endpoints = {}
+        matching_endpoints: ty.Dict[str, ty.List[discover.EndpointData]] = {}
 
         for service in self.normalize_catalog():
-
             if service_type and not discover._SERVICE_TYPES.is_match(
-                    service_type, service['type']):
+                service_type, service['type']
+            ):
                 continue
 
-            if (service_name and service['name'] and
-                    service_name != service['name']):
+            if (
+                service_name
+                and service['name']
+                and service_name != service['name']
+            ):
                 continue
 
-            if (service_id and service['id'] and
-                    service_id != service['id']):
+            if service_id and service['id'] and service_id != service['id']:
                 continue
 
             matching_endpoints.setdefault(service['type'], [])
@@ -198,27 +233,38 @@ class ServiceCatalog(metaclass=abc.ABCMeta):
                         interface=endpoint['interface'],
                         region_name=endpoint['region_name'],
                         endpoint_id=endpoint['id'],
-                        raw_endpoint=endpoint['raw_endpoint']))
+                        raw_endpoint=endpoint['raw_endpoint'],
+                    )
+                )
 
         if not interfaces:
             return self._endpoints_by_type(service_type, matching_endpoints)
 
-        ret = {}
+        ret: ty.Dict[str, ty.List[discover.EndpointData]] = {}
         for matched_service_type, endpoints in matching_endpoints.items():
             if not endpoints:
                 ret[matched_service_type] = []
                 continue
-            matches_by_interface = {}
+
+            matches_by_interface: ty.Dict[
+                str, ty.List[discover.EndpointData]
+            ] = {}
             for endpoint in endpoints:
                 matches_by_interface.setdefault(endpoint.interface, [])
                 matches_by_interface[endpoint.interface].append(endpoint)
-            best_interface = [i for i in interfaces
-                              if i in matches_by_interface.keys()][0]
+
+            best_interface = [
+                i for i in interfaces if i in matches_by_interface.keys()
+            ][0]
             ret[matched_service_type] = matches_by_interface[best_interface]
 
         return self._endpoints_by_type(service_type, ret)
 
-    def _endpoints_by_type(self, requested, endpoints):
+    def _endpoints_by_type(
+        self,
+        requested: ty.Optional[str],
+        endpoints: ty.Dict[str, ty.List[discover.EndpointData]],
+    ) -> ty.Dict[str, ty.List[discover.EndpointData]]:
         """Get the approrpriate endpoints from the list of given endpoints.
 
         Per the service type alias rules:
@@ -236,7 +282,7 @@ class ServiceCatalog(metaclass=abc.ABCMeta):
 
         :param str requested:
             The service_type as requested by the user.
-        :param dict sc:
+        :param dict endpoints:
             A dictionary keyed by found service_type. Values are opaque to
             this method.
 
@@ -279,9 +325,17 @@ class ServiceCatalog(metaclass=abc.ABCMeta):
         # part if we do. Raise this so that we can panic in unit tests.
         raise ValueError("Programming error choosing an endpoint.")
 
-    def get_endpoints(self, service_type=None, interface=None,
-                      region_name=None, service_name=None,
-                      service_id=None, endpoint_id=None):
+    def get_endpoints(
+        self,
+        service_type: ty.Optional[str] = None,
+        interface: ty.Union[
+            str, ty.List[str], ty.Tuple[str], ty.Set[str], None
+        ] = None,
+        region_name: ty.Optional[str] = None,
+        service_name: ty.Optional[str] = None,
+        service_id: ty.Optional[str] = None,
+        endpoint_id: ty.Optional[str] = None,
+    ) -> ty.Dict[str, ty.List[ty.Union[str, None]]]:
         """Fetch and filter endpoint data for the specified service(s).
 
         Returns endpoints for the specified service (or all) containing
@@ -294,17 +348,29 @@ class ServiceCatalog(metaclass=abc.ABCMeta):
         Returns a dict keyed by service_type with a list of endpoint dicts
         """
         endpoints_data = self.get_endpoints_data(
-            service_type=service_type, interface=interface,
-            region_name=region_name, service_name=service_name,
-            service_id=service_id, endpoint_id=endpoint_id)
+            service_type=service_type,
+            interface=interface,
+            region_name=region_name,
+            service_name=service_name,
+            service_id=service_id,
+            endpoint_id=endpoint_id,
+        )
         endpoints = {}
         for service_type, data in endpoints_data.items():
             endpoints[service_type] = self._denormalize_endpoints(data)
         return endpoints
 
-    def get_endpoint_data_list(self, service_type=None, interface='public',
-                               region_name=None, service_name=None,
-                               service_id=None, endpoint_id=None):
+    def get_endpoint_data_list(
+        self,
+        service_type: ty.Optional[str] = None,
+        interface: ty.Union[
+            str, ty.List[str], ty.Tuple[str], ty.Set[str], None
+        ] = 'public',
+        region_name: ty.Optional[str] = None,
+        service_name: ty.Optional[str] = None,
+        service_id: ty.Optional[str] = None,
+        endpoint_id: ty.Optional[str] = None,
+    ) -> ty.List[discover.EndpointData]:
         """Fetch a flat list of matching EndpointData objects.
 
         Fetch the endpoints from the service catalog for a particular
@@ -327,17 +393,27 @@ class ServiceCatalog(metaclass=abc.ABCMeta):
         :returns: a list of matching EndpointData objects
         :rtype: list(`keystoneauth1.discover.EndpointData`)
         """
-        endpoints = self.get_endpoints_data(service_type=service_type,
-                                            interface=interface,
-                                            region_name=region_name,
-                                            service_name=service_name,
-                                            service_id=service_id,
-                                            endpoint_id=endpoint_id)
+        endpoints = self.get_endpoints_data(
+            service_type=service_type,
+            interface=interface,
+            region_name=region_name,
+            service_name=service_name,
+            service_id=service_id,
+            endpoint_id=endpoint_id,
+        )
         return [endpoint for data in endpoints.values() for endpoint in data]
 
-    def get_urls(self, service_type=None, interface='public',
-                 region_name=None, service_name=None,
-                 service_id=None, endpoint_id=None):
+    def get_urls(
+        self,
+        service_type: ty.Optional[str] = None,
+        interface: ty.Union[
+            str, ty.List[str], ty.Tuple[str], ty.Set[str], None
+        ] = 'public',
+        region_name: ty.Optional[str] = None,
+        service_name: ty.Optional[str] = None,
+        service_id: ty.Optional[str] = None,
+        endpoint_id: ty.Optional[str] = None,
+    ) -> ty.Tuple[ty.Optional[str], ...]:
         """Fetch endpoint urls from the service catalog.
 
         Fetch the urls of endpoints from the service catalog for a particular
@@ -359,17 +435,27 @@ class ServiceCatalog(metaclass=abc.ABCMeta):
 
         :returns: tuple of urls
         """
-        endpoints = self.get_endpoint_data_list(service_type=service_type,
-                                                interface=interface,
-                                                region_name=region_name,
-                                                service_name=service_name,
-                                                service_id=service_id,
-                                                endpoint_id=endpoint_id)
+        endpoints = self.get_endpoint_data_list(
+            service_type=service_type,
+            interface=interface,
+            region_name=region_name,
+            service_name=service_name,
+            service_id=service_id,
+            endpoint_id=endpoint_id,
+        )
         return tuple([endpoint.url for endpoint in endpoints])
 
-    def url_for(self, service_type=None, interface='public',
-                region_name=None, service_name=None,
-                service_id=None, endpoint_id=None):
+    def url_for(
+        self,
+        service_type: ty.Optional[str] = None,
+        interface: ty.Union[
+            str, ty.List[str], ty.Tuple[str], ty.Set[str], None
+        ] = 'public',
+        region_name: ty.Optional[str] = None,
+        service_name: ty.Optional[str] = None,
+        service_id: ty.Optional[str] = None,
+        endpoint_id: ty.Optional[str] = None,
+    ) -> ty.Optional[str]:
         """Fetch an endpoint from the service catalog.
 
         Fetch the specified endpoint from the service catalog for
@@ -389,16 +475,26 @@ class ServiceCatalog(metaclass=abc.ABCMeta):
         :param string service_id: The identifier of a service.
         :param string endpoint_id: The identifier of an endpoint.
         """
-        return self.endpoint_data_for(service_type=service_type,
-                                      interface=interface,
-                                      region_name=region_name,
-                                      service_name=service_name,
-                                      service_id=service_id,
-                                      endpoint_id=endpoint_id).url
+        return self.endpoint_data_for(
+            service_type=service_type,
+            interface=interface,
+            region_name=region_name,
+            service_name=service_name,
+            service_id=service_id,
+            endpoint_id=endpoint_id,
+        ).url
 
-    def endpoint_data_for(self, service_type=None, interface='public',
-                          region_name=None, service_name=None,
-                          service_id=None, endpoint_id=None):
+    def endpoint_data_for(
+        self,
+        service_type: ty.Optional[str] = None,
+        interface: ty.Union[
+            str, ty.List[str], ty.Tuple[str], ty.Set[str], None
+        ] = 'public',
+        region_name: ty.Optional[str] = None,
+        service_name: ty.Optional[str] = None,
+        service_id: ty.Optional[str] = None,
+        endpoint_id: ty.Optional[str] = None,
+    ) -> discover.EndpointData:
         """Fetch endpoint data from the service catalog.
 
         Fetch the specified endpoint data from the service catalog for
@@ -427,34 +523,30 @@ class ServiceCatalog(metaclass=abc.ABCMeta):
             region_name=region_name,
             service_name=service_name,
             service_id=service_id,
-            endpoint_id=endpoint_id)
+            endpoint_id=endpoint_id,
+        )
 
         if endpoint_data_list:
             return endpoint_data_list[0]
 
         if service_name and region_name:
-            msg = ('%(interface)s endpoint for %(service_type)s service '
-                   'named %(service_name)s in %(region_name)s region not '
-                   'found' %
-                   {'interface': interface,
-                    'service_type': service_type, 'service_name': service_name,
-                    'region_name': region_name})
+            msg = (
+                f'{interface} endpoint for {service_type} service '
+                f'named {service_name} in {region_name} region not '
+                'found'
+            )
         elif service_name:
-            msg = ('%(interface)s endpoint for %(service_type)s service '
-                   'named %(service_name)s not found' %
-                   {'interface': interface,
-                    'service_type': service_type,
-                    'service_name': service_name})
+            msg = (
+                f'{interface} endpoint for {service_type} service '
+                f'named {service_name} not found'
+            )
         elif region_name:
-            msg = ('%(interface)s endpoint for %(service_type)s service '
-                   'in %(region_name)s region not found' %
-                   {'interface': interface,
-                    'service_type': service_type, 'region_name': region_name})
+            msg = (
+                f'{interface} endpoint for {service_type} service '
+                f'in {region_name} region not found'
+            )
         else:
-            msg = ('%(interface)s endpoint for %(service_type)s service '
-                   'not found' %
-                   {'interface': interface,
-                    'service_type': service_type})
+            msg = f'{interface} endpoint for {service_type} service not found'
 
         raise exceptions.EndpointNotFound(msg)
 
@@ -466,23 +558,27 @@ class ServiceCatalogV2(ServiceCatalog):
     """
 
     @classmethod
-    def from_token(cls, token):
+    def from_token(cls, token: ty.Dict[str, ty.Any]) -> ty_ext.Self:
         if 'access' not in token:
             raise ValueError('Invalid token format for fetching catalog')
 
         return cls(token['access'].get('serviceCatalog', {}))
 
     @staticmethod
-    def normalize_interface(interface):
+    def normalize_interface(interface: str) -> str:
         if interface and 'URL' not in interface:
             interface += 'URL'
 
         return interface
 
-    def is_interface_match(self, endpoint, interface):
+    def is_interface_match(
+        self, endpoint: ty.Dict[str, str], interface: str
+    ) -> bool:
         return interface in endpoint
 
-    def _normalize_endpoints(self, endpoints):
+    def _normalize_endpoints(
+        self, endpoints: ty.List[ty.Dict[str, ty.Any]]
+    ) -> ty.List[ty.Dict[str, ty.Any]]:
         """Translate endpoint description dicts into v3 form.
 
         Takes a raw endpoint description from the catalog and changes
@@ -498,8 +594,9 @@ class ServiceCatalogV2(ServiceCatalog):
         for endpoint in endpoints:
             raw_endpoint = endpoint.copy()
             interface_urls = {}
-            interface_keys = [key for key in endpoint.keys()
-                              if key.endswith('URL')]
+            interface_keys = [
+                key for key in endpoint.keys() if key.endswith('URL')
+            ]
             for key in interface_keys:
                 interface = self.normalize_interface(key)
                 interface_urls[interface] = endpoint.pop(key)
@@ -512,7 +609,9 @@ class ServiceCatalogV2(ServiceCatalog):
                 new_endpoints.append(new_endpoint)
         return new_endpoints
 
-    def _denormalize_endpoints(self, endpoints):
+    def _denormalize_endpoints(
+        self, endpoints: ty.List[discover.EndpointData]
+    ) -> ty.List[ty.Union[str, None]]:
         """Return original endpoint description dicts.
 
         Takes a list of EndpointData objects and returns the original
@@ -522,19 +621,18 @@ class ServiceCatalogV2(ServiceCatalog):
 
         :returns: List of endpoint description dicts in original catalog format
         """
-        raw_endpoints = super(ServiceCatalogV2, self)._denormalize_endpoints(
-            endpoints)
+        raw_endpoints = super()._denormalize_endpoints(endpoints)
         # The same raw endpoint content will be in the list once for each
         # v2 endpoint_type entry. We only need one of them in the resulting
         # list. So keep a list of the string versions.
         seen = {}
-        endpoints = []
+        ret = []
         for endpoint in raw_endpoints:
             if str(endpoint) in seen:
                 continue
             seen[str(endpoint)] = True
-            endpoints.append(endpoint)
-        return endpoints
+            ret.append(endpoint)
+        return ret
 
 
 class ServiceCatalogV3(ServiceCatalog):
@@ -544,20 +642,22 @@ class ServiceCatalogV3(ServiceCatalog):
     """
 
     @classmethod
-    def from_token(cls, token):
+    def from_token(cls, token: ty.Dict[str, ty.Any]) -> ty_ext.Self:
         if 'token' not in token:
             raise ValueError('Invalid token format for fetching catalog')
 
         return cls(token['token'].get('catalog', {}))
 
     @staticmethod
-    def normalize_interface(interface):
+    def normalize_interface(interface: str) -> str:
         if interface:
             interface = interface.rstrip('URL')
 
         return interface
 
-    def is_interface_match(self, endpoint, interface):
+    def is_interface_match(
+        self, endpoint: ty.Dict[str, str], interface: str
+    ) -> bool:
         try:
             return interface == endpoint['interface']
         except KeyError:
