@@ -10,14 +10,24 @@
 # License for the specific language governing permissions and limitations
 # under the License.
 
+import argparse
+import collections
+import logging
 import os
+import typing as ty
 import warnings
 
+import requests
+
 from keystoneauth1 import _fair_semaphore
+from keystoneauth1 import discover
 from keystoneauth1 import session
 
+if ty.TYPE_CHECKING:
+    from keystoneauth1 import plugin
 
-class Adapter(object):
+
+class _BaseAdapter:
     """An instance of a session with local variables.
 
     A session is a global object that is shared around amongst many clients. It
@@ -113,26 +123,46 @@ class Adapter(object):
         a maximum of 60 seconds is used.
     """
 
-    client_name = None
-    client_version = None
+    client_name: ty.Optional[str] = None
+    client_version: ty.Optional[str] = None
 
-    def __init__(self, session, service_type=None, service_name=None,
-                 interface=None, region_name=None, endpoint_override=None,
-                 version=None, auth=None, user_agent=None,
-                 connect_retries=None, logger=None, allow=None,
-                 additional_headers=None, client_name=None,
-                 client_version=None, allow_version_hack=None,
-                 global_request_id=None,
-                 min_version=None, max_version=None,
-                 default_microversion=None, status_code_retries=None,
-                 retriable_status_codes=None, raise_exc=None,
-                 rate_limit=None, concurrency=None,
-                 connect_retry_delay=None, status_code_retry_delay=None,
-                 ):
+    def __init__(
+        self,
+        session: session.Session,
+        service_type: ty.Optional[str] = None,
+        service_name: ty.Optional[str] = None,
+        interface: ty.Optional[str] = None,
+        region_name: ty.Optional[str] = None,
+        endpoint_override: ty.Optional[str] = None,
+        version: ty.Optional[str] = None,
+        auth: ty.Optional['plugin.BaseAuthPlugin'] = None,
+        user_agent: ty.Optional[str] = None,
+        connect_retries: ty.Optional[int] = None,
+        logger: ty.Optional[logging.Logger] = None,
+        allow: ty.Optional[ty.Dict[str, ty.Any]] = None,
+        additional_headers: ty.Optional[
+            collections.abc.MutableMapping[str, str]
+        ] = None,
+        client_name: ty.Optional[str] = None,
+        client_version: ty.Optional[str] = None,
+        allow_version_hack: ty.Optional[bool] = None,
+        global_request_id: ty.Optional[str] = None,
+        min_version: ty.Optional[str] = None,
+        max_version: ty.Optional[str] = None,
+        default_microversion: ty.Optional[str] = None,
+        status_code_retries: ty.Optional[int] = None,
+        retriable_status_codes: ty.Optional[ty.List[int]] = None,
+        raise_exc: ty.Optional[bool] = None,
+        rate_limit: ty.Optional[float] = None,
+        concurrency: ty.Optional[int] = None,
+        connect_retry_delay: ty.Optional[float] = None,
+        status_code_retry_delay: ty.Optional[float] = None,
+    ):
         if version and (min_version or max_version):
             raise TypeError(
                 "version is mutually exclusive with min_version and"
-                " max_version")
+                " max_version"
+            )
         # NOTE(jamielennox): when adding new parameters to adapter please also
         # add them to the adapter call in httpclient.HTTPClient.__init__ as
         # well as to load_adapter_from_argparse below if the argument is
@@ -175,9 +205,12 @@ class Adapter(object):
             rate_delay = 1.0 / rate_limit
 
         self._rate_semaphore = _fair_semaphore.FairSemaphore(
-            concurrency, rate_delay)
+            concurrency, rate_delay
+        )
 
-    def _set_endpoint_filter_kwargs(self, kwargs):
+    def _set_endpoint_filter_kwargs(
+        self, kwargs: ty.Dict[str, object]
+    ) -> None:
         if self.service_type:
             kwargs.setdefault('service_type', self.service_type)
         if self.service_name:
@@ -195,10 +228,20 @@ class Adapter(object):
         if self.allow_version_hack is not None:
             kwargs.setdefault('allow_version_hack', self.allow_version_hack)
 
-    def request(self, url, method, **kwargs):
+    def _request(
+        self, url: str, method: str, **kwargs: ty.Any
+    ) -> requests.Response:
         endpoint_filter = kwargs.setdefault('endpoint_filter', {})
         self._set_endpoint_filter_kwargs(endpoint_filter)
-
+        # NOTE(gmann): Convert or initlize the headers to
+        # CaseInsensitiveDict to make sure headers are
+        # case insensitive.
+        if kwargs.get('headers'):
+            kwargs['headers'] = requests.structures.CaseInsensitiveDict(
+                kwargs['headers']
+            )
+        else:
+            kwargs['headers'] = requests.structures.CaseInsensitiveDict()
         if self.endpoint_override:
             kwargs.setdefault('endpoint_override', self.endpoint_override)
 
@@ -206,13 +249,15 @@ class Adapter(object):
             kwargs.setdefault('auth', self.auth)
         if self.user_agent:
             kwargs.setdefault('user_agent', self.user_agent)
-        for arg in ('connect_retries', 'status_code_retries',
-                    'connect_retry_delay', 'status_code_retry_delay'):
+        for arg in (
+            'connect_retries',
+            'status_code_retries',
+            'connect_retry_delay',
+            'status_code_retry_delay',
+            'retriable_status_codes',
+        ):
             if getattr(self, arg) is not None:
                 kwargs.setdefault(arg, getattr(self, arg))
-        if self.retriable_status_codes:
-            kwargs.setdefault('retriable_status_codes',
-                              self.retriable_status_codes)
         if self.logger:
             kwargs.setdefault('logger', self.logger)
         if self.allow:
@@ -232,22 +277,27 @@ class Adapter(object):
                 kwargs.setdefault('rate_semaphore', self._rate_semaphore)
 
         else:
-            warnings.warn('Using keystoneclient sessions has been deprecated. '
-                          'Please update your software to use keystoneauth1.')
+            warnings.warn(
+                'Using keystoneclient sessions has been deprecated. '
+                'Please update your software to use keystoneauth1.'
+            )
 
         for k, v in self.additional_headers.items():
             kwargs.setdefault('headers', {}).setdefault(k, v)
 
         if self.global_request_id is not None:
             kwargs.setdefault('headers', {}).setdefault(
-                "X-OpenStack-Request-ID", self.global_request_id)
+                "X-OpenStack-Request-ID", self.global_request_id
+            )
 
         if self.raise_exc is not None:
             kwargs.setdefault('raise_exc', self.raise_exc)
 
         return self.session.request(url, method, **kwargs)
 
-    def get_token(self, auth=None):
+    def get_token(
+        self, auth: ty.Optional['plugin.BaseAuthPlugin'] = None
+    ) -> ty.Optional[str]:
         """Return a token as provided by the auth plugin.
 
         :param auth: The auth plugin to use for token. Overrides the plugin
@@ -262,7 +312,11 @@ class Adapter(object):
         """
         return self.session.get_token(auth or self.auth)
 
-    def get_endpoint(self, auth=None, **kwargs):
+    def get_endpoint(
+        self,
+        auth: ty.Optional['plugin.BaseAuthPlugin'] = None,
+        **kwargs: ty.Any,
+    ) -> ty.Optional[str]:
         """Get an endpoint as provided by the auth plugin.
 
         :param auth: The auth plugin to use for token. Overrides the plugin on
@@ -281,7 +335,9 @@ class Adapter(object):
         self._set_endpoint_filter_kwargs(kwargs)
         return self.session.get_endpoint(auth or self.auth, **kwargs)
 
-    def get_endpoint_data(self, auth=None):
+    def get_endpoint_data(
+        self, auth: ty.Optional['plugin.BaseAuthPlugin'] = None
+    ) -> ty.Optional['discover.EndpointData']:
         """Get the endpoint data for this Adapter's endpoint.
 
         :param auth: The auth plugin to use for token. Overrides the plugin on
@@ -295,7 +351,7 @@ class Adapter(object):
         :returns: Endpoint data if available or None.
         :rtype: keystoneauth1.discover.EndpointData
         """
-        kwargs = {}
+        kwargs: ty.Dict[str, ty.Any] = {}
         self._set_endpoint_filter_kwargs(kwargs)
         if self.endpoint_override:
             kwargs['endpoint_override'] = self.endpoint_override
@@ -303,9 +359,10 @@ class Adapter(object):
         return self.session.get_endpoint_data(auth or self.auth, **kwargs)
 
     def get_all_version_data(
-            self,
-            interface='public',
-            region_name=None):
+        self, interface: str = 'public', region_name: ty.Optional[str] = None
+    ) -> ty.Dict[
+        str, ty.Dict[str, ty.Dict[str, ty.List[discover.VersionData]]]
+    ]:
         """Get data about all versions of a service.
 
         :param interface:
@@ -323,9 +380,14 @@ class Adapter(object):
         return self.session.get_all_version_data(
             interface=interface,
             region_name=region_name,
-            service_type=self.service_type)
+            service_type=self.service_type,
+        )
 
-    def get_api_major_version(self, auth=None, **kwargs):
+    def get_api_major_version(
+        self,
+        auth: ty.Optional['plugin.BaseAuthPlugin'] = None,
+        **kwargs: ty.Any,
+    ) -> ty.Optional[ty.Tuple[ty.Union[int, float], ...]]:
         """Get the major API version as provided by the auth plugin.
 
         :param auth: The auth plugin to use for token. Overrides the plugin on
@@ -344,11 +406,20 @@ class Adapter(object):
 
         return self.session.get_api_major_version(auth or self.auth, **kwargs)
 
-    def invalidate(self, auth=None):
-        """Invalidate an authentication plugin."""
+    def invalidate(
+        self, auth: ty.Optional['plugin.BaseAuthPlugin'] = None
+    ) -> bool:
+        """Invalidate an authentication plugin.
+
+        :param auth: The auth plugin to invalidate. Overrides the plugin on the
+                     session. (optional)
+        :type auth: keystoneauth1.plugin.BaseAuthPlugin
+        """
         return self.session.invalidate(auth or self.auth)
 
-    def get_user_id(self, auth=None):
+    def get_user_id(
+        self, auth: ty.Optional['plugin.BaseAuthPlugin'] = None
+    ) -> ty.Optional[str]:
         """Return the authenticated user_id as provided by the auth plugin.
 
         :param auth: The auth plugin to use for token. Overrides the plugin
@@ -365,7 +436,9 @@ class Adapter(object):
         """
         return self.session.get_user_id(auth or self.auth)
 
-    def get_project_id(self, auth=None):
+    def get_project_id(
+        self, auth: ty.Optional['plugin.BaseAuthPlugin'] = None
+    ) -> ty.Optional[str]:
         """Return the authenticated project_id as provided by the auth plugin.
 
         :param auth: The auth plugin to use for token. Overrides the plugin
@@ -382,27 +455,13 @@ class Adapter(object):
         """
         return self.session.get_project_id(auth or self.auth)
 
-    def get(self, url, **kwargs):
-        return self.request(url, 'GET', **kwargs)
-
-    def head(self, url, **kwargs):
-        return self.request(url, 'HEAD', **kwargs)
-
-    def post(self, url, **kwargs):
-        return self.request(url, 'POST', **kwargs)
-
-    def put(self, url, **kwargs):
-        return self.request(url, 'PUT', **kwargs)
-
-    def patch(self, url, **kwargs):
-        return self.request(url, 'PATCH', **kwargs)
-
-    def delete(self, url, **kwargs):
-        return self.request(url, 'DELETE', **kwargs)
-
     # TODO(efried): Move this to loading.adapter.Adapter
     @classmethod
-    def register_argparse_arguments(cls, parser, service_type=None):
+    def register_argparse_arguments(
+        cls,
+        parser: argparse.ArgumentParser,
+        service_type: ty.Optional[str] = None,
+    ) -> None:
         """Attach arguments to a given argparse Parser for Adapters.
 
         :param parser: The argparse parser to attach options to.
@@ -411,48 +470,57 @@ class Adapter(object):
         """
         adapter_group = parser.add_argument_group(
             'Service Options',
-            'Options controlling the specialization of the API'
-            ' Connection from information found in the catalog')
+            'Options controlling the specialization of the API '
+            'Connection from information found in the catalog',
+        )
 
         adapter_group.add_argument(
             '--os-service-type',
             metavar='<name>',
             default=os.environ.get('OS_SERVICE_TYPE', service_type),
-            help='Service type to request from the catalog')
+            help='Service type to request from the catalog',
+        )
 
         adapter_group.add_argument(
             '--os-service-name',
             metavar='<name>',
             default=os.environ.get('OS_SERVICE_NAME', None),
-            help='Service name to request from the catalog')
+            help='Service name to request from the catalog',
+        )
 
         adapter_group.add_argument(
             '--os-interface',
             metavar='<name>',
             default=os.environ.get('OS_INTERFACE', 'public'),
-            help='API Interface to use [public, internal, admin]')
+            help='API Interface to use [public, internal, admin]',
+        )
 
         adapter_group.add_argument(
             '--os-region-name',
             metavar='<name>',
             default=os.environ.get('OS_REGION_NAME', None),
-            help='Region of the cloud to use')
+            help='Region of the cloud to use',
+        )
 
         adapter_group.add_argument(
             '--os-endpoint-override',
             metavar='<name>',
             default=os.environ.get('OS_ENDPOINT_OVERRIDE', None),
-            help='Endpoint to use instead of the endpoint in the catalog')
+            help='Endpoint to use instead of the endpoint in the catalog',
+        )
 
         adapter_group.add_argument(
             '--os-api-version',
             metavar='<name>',
             default=os.environ.get('OS_API_VERSION', None),
-            help='Which version of the service API to use')
+            help='Which version of the service API to use',
+        )
 
     # TODO(efried): Move this to loading.adapter.Adapter
     @classmethod
-    def register_service_argparse_arguments(cls, parser, service_type):
+    def register_service_argparse_arguments(
+        cls, parser: argparse.ArgumentParser, service_type: str
+    ) -> None:
         """Attach arguments to a given argparse Parser for Adapters.
 
         :param parser: The argparse parser to attach options to.
@@ -462,69 +530,115 @@ class Adapter(object):
         """
         service_env = service_type.upper().replace('-', '_')
         adapter_group = parser.add_argument_group(
-            '{service_type} Service Options'.format(
-                service_type=service_type.title()),
-            'Options controlling the specialization of the {service_type}'
-            ' API Connection from information found in the catalog'.format(
-                service_type=service_type.title()))
+            f'{service_type.title()} Service Options',
+            f'Options controlling the specialization of the '
+            f'{service_type.title()} API Connection from information found in '
+            f'the catalog',
+        )
 
         adapter_group.add_argument(
-            '--os-{service_type}-service-type'.format(
-                service_type=service_type),
+            f'--os-{service_type}-service-type',
             metavar='<name>',
-            default=os.environ.get(
-                'OS_{service_type}_SERVICE_TYPE'.format(
-                    service_type=service_env), None),
-            help=('Service type to request from the catalog for the'
-                  ' {service_type} service'.format(
-                      service_type=service_type)))
+            default=os.environ.get(f'OS_{service_env}_SERVICE_TYPE', None),
+            help=(
+                f'Service type to request from the catalog for the '
+                f'{service_type} service'
+            ),
+        )
 
         adapter_group.add_argument(
-            '--os-{service_type}-service-name'.format(
-                service_type=service_type),
+            f'--os-{service_type}-service-name',
             metavar='<name>',
-            default=os.environ.get(
-                'OS_{service_type}_SERVICE_NAME'.format(
-                    service_type=service_env), None),
-            help=('Service name to request from the catalog for the'
-                  ' {service_type} service'.format(
-                      service_type=service_type)))
+            default=os.environ.get(f'OS_{service_env}_SERVICE_NAME', None),
+            help=(
+                f'Service name to request from the catalog for the '
+                f'{service_type} service'
+            ),
+        )
 
         adapter_group.add_argument(
-            '--os-{service_type}-interface'.format(
-                service_type=service_type),
+            f'--os-{service_type}-interface',
             metavar='<name>',
-            default=os.environ.get(
-                'OS_{service_type}_INTERFACE'.format(
-                    service_type=service_env), None),
-            help=('API Interface to use for the {service_type} service'
-                  ' [public, internal, admin]'.format(
-                      service_type=service_type)))
+            default=os.environ.get(f'OS_{service_env}_INTERFACE', None),
+            help=(
+                f'API Interface to use for the {service_type} service '
+                f'[public, internal, admin]'
+            ),
+        )
 
         adapter_group.add_argument(
-            '--os-{service_type}-api-version'.format(
-                service_type=service_type),
+            f'--os-{service_type}-api-version',
             metavar='<name>',
-            default=os.environ.get(
-                'OS_{service_type}_API_VERSION'.format(
-                    service_type=service_env), None),
-            help=('Which version of the service API to use for'
-                  ' the {service_type} service'.format(
-                      service_type=service_type)))
+            default=os.environ.get(f'OS_{service_env}_API_VERSION', None),
+            help=(
+                f'Which version of the service API to use for '
+                f'the {service_type} service'
+            ),
+        )
 
         adapter_group.add_argument(
-            '--os-{service_type}-endpoint-override'.format(
-                service_type=service_type),
+            f'--os-{service_type}-endpoint-override',
             metavar='<name>',
             default=os.environ.get(
-                'OS_{service_type}_ENDPOINT_OVERRIDE'.format(
-                    service_type=service_env), None),
-            help=('Endpoint to use for the {service_type} service'
-                  ' instead of the endpoint in the catalog'.format(
-                      service_type=service_type)))
+                f'OS_{service_env}_ENDPOINT_OVERRIDE', None
+            ),
+            help=(
+                f'Endpoint to use for the {service_type} service '
+                f'instead of the endpoint in the catalog'
+            ),
+        )
 
 
-class LegacyJsonAdapter(Adapter):
+class Adapter(_BaseAdapter):
+    def request(
+        self, url: str, method: str, **kwargs: ty.Any
+    ) -> requests.Response:
+        return self._request(url, method, **kwargs)
+
+    def get(self, url: str, **kwargs: ty.Any) -> requests.Response:
+        """Perform a GET request.
+
+        This calls :py:meth:`.request()` with ``method`` set to ``GET``.
+        """
+        return self.request(url, 'GET', **kwargs)
+
+    def head(self, url: str, **kwargs: ty.Any) -> requests.Response:
+        """Perform a HEAD request.
+
+        This calls :py:meth:`.request()` with ``method`` set to ``HEAD``.
+        """
+        return self.request(url, 'HEAD', **kwargs)
+
+    def post(self, url: str, **kwargs: ty.Any) -> requests.Response:
+        """Perform a POST request.
+
+        This calls :py:meth:`.request()` with ``method`` set to ``POST``.
+        """
+        return self.request(url, 'POST', **kwargs)
+
+    def put(self, url: str, **kwargs: ty.Any) -> requests.Response:
+        """Perform a PUT request.
+
+        This calls :py:meth:`.request()` with ``method`` set to ``PUT``.
+        """
+        return self.request(url, 'PUT', **kwargs)
+
+    def patch(self, url: str, **kwargs: ty.Any) -> requests.Response:
+        """Perform a PATCH request.
+
+        This calls :py:meth:`.request()` with ``method`` set to ``PATCH``.
+        """
+        return self.request(url, 'PATCH', **kwargs)
+
+    def delete(self, url: str, **kwargs: ty.Any) -> requests.Response:
+        """Perform a DELETE request.
+
+        This calls :py:meth:`.request()` with ``method`` set to ``DELETE``.
+        """
+        return self.request(url, 'DELETE', **kwargs)
+
+
+class LegacyJsonAdapter(_BaseAdapter):
     """Make something that looks like an old HTTPClient.
 
     A common case when using an adapter is that we want an interface similar to
@@ -533,7 +647,9 @@ class LegacyJsonAdapter(Adapter):
     You probably don't want this if you are starting from scratch.
     """
 
-    def request(self, *args, **kwargs):
+    def request(
+        self, url: str, method: str, **kwargs: ty.Any
+    ) -> ty.Tuple[requests.Response, object]:
         headers = kwargs.setdefault('headers', {})
         headers.setdefault('Accept', 'application/json')
 
@@ -542,7 +658,7 @@ class LegacyJsonAdapter(Adapter):
         except KeyError:
             pass
 
-        resp = super(LegacyJsonAdapter, self).request(*args, **kwargs)
+        resp = self._request(url, method, **kwargs)
 
         try:
             body = resp.json()
@@ -551,14 +667,76 @@ class LegacyJsonAdapter(Adapter):
 
         return resp, body
 
+    def get(
+        self, url: str, **kwargs: ty.Any
+    ) -> ty.Tuple[requests.Response, object]:
+        """Perform a GET request.
+
+        This calls :py:meth:`.request()` with ``method`` set to ``GET``.
+        """
+        return self.request(url, 'GET', **kwargs)
+
+    def head(
+        self, url: str, **kwargs: ty.Any
+    ) -> ty.Tuple[requests.Response, object]:
+        """Perform a HEAD request.
+
+        This calls :py:meth:`.request()` with ``method`` set to ``HEAD``.
+        """
+        return self.request(url, 'HEAD', **kwargs)
+
+    def post(
+        self, url: str, **kwargs: ty.Any
+    ) -> ty.Tuple[requests.Response, object]:
+        """Perform a POST request.
+
+        This calls :py:meth:`.request()` with ``method`` set to ``POST``.
+        """
+        return self.request(url, 'POST', **kwargs)
+
+    def put(
+        self, url: str, **kwargs: ty.Any
+    ) -> ty.Tuple[requests.Response, object]:
+        """Perform a PUT request.
+
+        This calls :py:meth:`.request()` with ``method`` set to ``PUT``.
+        """
+        return self.request(url, 'PUT', **kwargs)
+
+    def patch(
+        self, url: str, **kwargs: ty.Any
+    ) -> ty.Tuple[requests.Response, object]:
+        """Perform a PATCH request.
+
+        This calls :py:meth:`.request()` with ``method`` set to ``PATCH``.
+        """
+        return self.request(url, 'PATCH', **kwargs)
+
+    def delete(
+        self, url: str, **kwargs: ty.Any
+    ) -> ty.Tuple[requests.Response, object]:
+        """Perform a DELETE request.
+
+        This calls :py:meth:`.request()` with ``method`` set to ``DELETE``.
+        """
+        return self.request(url, 'DELETE', **kwargs)
+
 
 # TODO(efried): Deprecate this in favor of
 #               loading.adapter.register_argparse_arguments
-def register_adapter_argparse_arguments(*args, **kwargs):
-    return Adapter.register_argparse_arguments(*args, **kwargs)
+def register_adapter_argparse_arguments(
+    parser: argparse.ArgumentParser, service_type: ty.Optional[str] = None
+) -> None:
+    return Adapter.register_argparse_arguments(
+        parser=parser, service_type=service_type
+    )
 
 
 # TODO(efried): Deprecate this in favor of
 #               loading.adapter.register_service_argparse_arguments
-def register_service_adapter_argparse_arguments(*args, **kwargs):
-    return Adapter.register_service_argparse_arguments(*args, **kwargs)
+def register_service_adapter_argparse_arguments(
+    parser: argparse.ArgumentParser, service_type: str
+) -> None:
+    return Adapter.register_service_argparse_arguments(
+        parser=parser, service_type=service_type
+    )
