@@ -16,28 +16,31 @@ import functools
 import hashlib
 import json
 import threading
-
-import six
+import typing as ty
 
 from keystoneauth1 import _utils as utils
 from keystoneauth1 import access
 from keystoneauth1 import discover
 from keystoneauth1 import exceptions
 from keystoneauth1 import plugin
+from keystoneauth1 import session as ks_session
 
 LOG = utils.get_logger(__name__)
 
 
-@six.add_metaclass(abc.ABCMeta)
-class BaseIdentityPlugin(plugin.BaseAuthPlugin):
-
+class BaseIdentityPlugin(plugin.BaseAuthPlugin, metaclass=abc.ABCMeta):
     # we count a token as valid (not needing refreshing) if it is valid for at
     # least this many seconds before the token expiry time
-    MIN_TOKEN_LIFE_SECONDS = 120
+    MIN_TOKEN_LIFE_SECONDS: int = 120
 
-    def __init__(self, auth_url=None, reauthenticate=True):
+    auth_url: ty.Optional[str]
+    auth_ref: ty.Optional[access.AccessInfo]
+    reauthenticate: bool
 
-        super(BaseIdentityPlugin, self).__init__()
+    def __init__(
+        self, auth_url: ty.Optional[str] = None, reauthenticate: bool = True
+    ):
+        super().__init__()
 
         self.auth_url = auth_url
         self.auth_ref = None
@@ -46,7 +49,9 @@ class BaseIdentityPlugin(plugin.BaseAuthPlugin):
         self._lock = threading.Lock()
 
     @abc.abstractmethod
-    def get_auth_ref(self, session, **kwargs):
+    def get_auth_ref(
+        self, session: ks_session.Session, **kwargs: ty.Any
+    ) -> access.AccessInfo:
         """Obtain a token from an OpenStack Identity Service.
 
         This method is overridden by the various token version plugins.
@@ -71,7 +76,9 @@ class BaseIdentityPlugin(plugin.BaseAuthPlugin):
         :rtype: :class:`keystoneauth1.access.AccessInfo`
         """
 
-    def get_token(self, session, **kwargs):
+    def get_token(
+        self, session: ks_session.Session, **kwargs: ty.Any
+    ) -> ty.Optional[str]:
         """Return a valid auth token.
 
         If a valid token is not present then a new one will be fetched.
@@ -87,7 +94,7 @@ class BaseIdentityPlugin(plugin.BaseAuthPlugin):
         """
         return self.get_access(session).auth_token
 
-    def _needs_reauthenticate(self):
+    def _needs_reauthenticate(self) -> bool:
         """Return if the existing token needs to be re-authenticated.
 
         The token should be refreshed if it is about to expire.
@@ -109,7 +116,9 @@ class BaseIdentityPlugin(plugin.BaseAuthPlugin):
         # otherwise it's fine and use the existing one.
         return False
 
-    def get_access(self, session, **kwargs):
+    def get_access(
+        self, session: ks_session.Session, **kwargs: ty.Any
+    ) -> access.AccessInfo:
         """Fetch or return a current AccessInfo object.
 
         If a valid AccessInfo is present then it is returned otherwise a new
@@ -133,9 +142,12 @@ class BaseIdentityPlugin(plugin.BaseAuthPlugin):
             if self._needs_reauthenticate():
                 self.auth_ref = self.get_auth_ref(session)
 
+        # narrow type
+        assert self.auth_ref is not None  # nosec B101
+
         return self.auth_ref
 
-    def invalidate(self):
+    def invalidate(self) -> bool:
         """Invalidate the current authentication data.
 
         This should result in fetching a new token on next call.
@@ -155,11 +167,23 @@ class BaseIdentityPlugin(plugin.BaseAuthPlugin):
 
         return False
 
-    def get_endpoint_data(self, session, service_type=None, interface=None,
-                          region_name=None, service_name=None, allow=None,
-                          allow_version_hack=True, discover_versions=True,
-                          skip_discovery=False, min_version=None,
-                          max_version=None, endpoint_override=None, **kwargs):
+    def get_endpoint_data(
+        self,
+        session: ks_session.Session,
+        *,
+        endpoint_override: ty.Optional[str] = None,
+        discover_versions: bool = True,
+        service_type: ty.Optional[str] = None,
+        interface: ty.Optional[str] = None,
+        region_name: ty.Optional[str] = None,
+        service_name: ty.Optional[str] = None,
+        allow: ty.Optional[ty.Dict[str, ty.Any]] = None,
+        allow_version_hack: bool = True,
+        skip_discovery: bool = False,
+        min_version: ty.Optional[discover._RAW_VERSION_T] = None,
+        max_version: ty.Optional[discover._RAW_VERSION_T] = None,
+        **kwargs: ty.Any,
+    ) -> ty.Optional[discover.EndpointData]:
         """Return a valid endpoint data for a service.
 
         If a valid token is not present then a new one will be fetched using
@@ -174,6 +198,16 @@ class BaseIdentityPlugin(plugin.BaseAuthPlugin):
 
         :param session: A session object that can be used for communication.
         :type session: keystoneauth1.session.Session
+        :param str endpoint_override: URL to use instead of looking in the
+                                      catalog. Catalog lookup will be skipped,
+                                      but version discovery will be run.
+                                      Sets allow_version_hack to False
+                                      (optional)
+        :param bool discover_versions: Whether to get version metadata from
+                                       the version discovery document even
+                                       if it's not neccessary to fulfill the
+                                       major version request. (optional,
+                                       defaults to True)
         :param string service_type: The type of service to lookup the endpoint
                                     for. This plugin will return None (failure)
                                     if service_type is not provided.
@@ -192,11 +226,6 @@ class BaseIdentityPlugin(plugin.BaseAuthPlugin):
         :param bool allow_version_hack: Allow keystoneauth to hack up catalog
                                         URLS to support older schemes.
                                         (optional, default True)
-        :param bool discover_versions: Whether to get version metadata from
-                                       the version discovery document even
-                                       if it's not neccessary to fulfill the
-                                       major version request. (optional,
-                                       defaults to True)
         :param bool skip_discovery: Whether to skip version discovery even
                                     if a version has been given. This is useful
                                     if endpoint_override or similar has been
@@ -210,11 +239,6 @@ class BaseIdentityPlugin(plugin.BaseAuthPlugin):
                             exclusive with version. If min_version is given
                             with no max_version it is as if max version is
                             'latest'. (optional)
-        :param str endpoint_override: URL to use instead of looking in the
-                                      catalog. Catalog lookup will be skipped,
-                                      but version discovery will be run.
-                                      Sets allow_version_hack to False
-                                      (optional)
         :param kwargs: Ignored.
 
         :raises keystoneauth1.exceptions.http.HttpError: An error from an
@@ -226,7 +250,8 @@ class BaseIdentityPlugin(plugin.BaseAuthPlugin):
         allow = allow or {}
 
         min_version, max_version = discover._normalize_version_args(
-            None, min_version, max_version)
+            None, min_version, max_version, service_type=service_type
+        )
 
         # NOTE(jamielennox): if you specifically ask for requests to be sent to
         # the auth url then we can ignore many of the checks. Typically if you
@@ -236,7 +261,8 @@ class BaseIdentityPlugin(plugin.BaseAuthPlugin):
         if interface is plugin.AUTH_INTERFACE:
             endpoint_data = discover.EndpointData(
                 service_url=self.auth_url,
-                service_type=service_type or 'identity')
+                service_type=service_type or 'identity',
+            )
             project_id = None
         elif endpoint_override:
             # TODO(mordred) Make a code path that will look for a
@@ -245,10 +271,12 @@ class BaseIdentityPlugin(plugin.BaseAuthPlugin):
             #               For now, just use any information the use has
             #               provided.
             endpoint_data = discover.EndpointData(
+                service_url=endpoint_override,
                 catalog_url=endpoint_override,
                 interface=interface,
                 region_name=region_name,
-                service_name=service_name)
+                service_name=service_name,
+            )
             # Setting an endpoint_override then calling get_endpoint_data means
             # you absolutely want the discovery info for the URL in question.
             # There are no code flows where this will happen for any other
@@ -257,9 +285,11 @@ class BaseIdentityPlugin(plugin.BaseAuthPlugin):
             project_id = self.get_project_id(session)
         else:
             if not service_type:
-                LOG.warning('Plugin cannot return an endpoint without '
-                            'knowing the service type that is required. Add '
-                            'service_type to endpoint filtering data.')
+                LOG.warning(
+                    'Plugin cannot return an endpoint without '
+                    'knowing the service type that is required. Add '
+                    'service_type to endpoint filtering data.'
+                )
                 return None
 
             # It's possible for things higher in the stack, because of
@@ -275,7 +305,8 @@ class BaseIdentityPlugin(plugin.BaseAuthPlugin):
                 service_type=service_type,
                 interface=interface,
                 region_name=region_name,
-                service_name=service_name)
+                service_name=service_name,
+            )
             if not endpoint_data:
                 return None
 
@@ -290,10 +321,14 @@ class BaseIdentityPlugin(plugin.BaseAuthPlugin):
                 max_version=max_version,
                 cache=self._discovery_cache,
                 discover_versions=discover_versions,
-                allow_version_hack=allow_version_hack, allow=allow)
-        except (exceptions.DiscoveryFailure,
-                exceptions.HttpError,
-                exceptions.ConnectionError):
+                allow_version_hack=allow_version_hack,
+                allow=allow,
+            )
+        except (
+            exceptions.DiscoveryFailure,
+            exceptions.HttpError,
+            exceptions.ConnectionError,
+        ):
             # If a version was requested, we didn't find it, return
             # None.
             if max_version or min_version:
@@ -302,12 +337,21 @@ class BaseIdentityPlugin(plugin.BaseAuthPlugin):
             # should be fine
             return endpoint_data
 
-    def get_endpoint(self, session, service_type=None, interface=None,
-                     region_name=None, service_name=None, version=None,
-                     allow=None, allow_version_hack=True,
-                     skip_discovery=False,
-                     min_version=None, max_version=None,
-                     **kwargs):
+    def get_endpoint(
+        self,
+        session: ks_session.Session,
+        service_type: ty.Optional[str] = None,
+        interface: ty.Optional[str] = None,
+        region_name: ty.Optional[str] = None,
+        service_name: ty.Optional[str] = None,
+        version: ty.Optional[discover._RAW_VERSION_T] = None,
+        allow: ty.Optional[ty.Dict[str, ty.Any]] = None,
+        allow_version_hack: bool = True,
+        skip_discovery: bool = False,
+        min_version: ty.Optional[discover._RAW_VERSION_T] = None,
+        max_version: ty.Optional[discover._RAW_VERSION_T] = None,
+        **kwargs: ty.Any,
+    ) -> ty.Optional[str]:
         """Return a valid endpoint for a service.
 
         If a valid token is not present then a new one will be fetched using
@@ -365,26 +409,47 @@ class BaseIdentityPlugin(plugin.BaseAuthPlugin):
         # Explode `version` into min_version and max_version - everything below
         # here uses the latter rather than the former.
         min_version, max_version = discover._normalize_version_args(
-            version, min_version, max_version)
+            version, min_version, max_version, service_type=service_type
+        )
         # Set discover_versions to False since we're only going to return
         # a URL. Fetching the microversion data would be needlessly
         # expensive in the common case. However, discover_versions=False
         # will still run discovery if the version requested is not the
         # version in the catalog.
         endpoint_data = self.get_endpoint_data(
-            session, service_type=service_type, interface=interface,
-            region_name=region_name, service_name=service_name,
-            allow=allow, min_version=min_version, max_version=max_version,
-            discover_versions=False, skip_discovery=skip_discovery,
-            allow_version_hack=allow_version_hack, **kwargs)
+            session,
+            service_type=service_type,
+            interface=interface,
+            region_name=region_name,
+            service_name=service_name,
+            allow=allow,
+            min_version=min_version,
+            max_version=max_version,
+            discover_versions=False,
+            skip_discovery=skip_discovery,
+            allow_version_hack=allow_version_hack,
+            **kwargs,
+        )
         return endpoint_data.url if endpoint_data else None
 
-    def get_api_major_version(self, session, service_type=None, interface=None,
-                              region_name=None, service_name=None,
-                              version=None, allow=None,
-                              allow_version_hack=True, skip_discovery=False,
-                              discover_versions=False, min_version=None,
-                              max_version=None, **kwargs):
+    def get_api_major_version(
+        self,
+        session: ks_session.Session,
+        *,
+        endpoint_override: ty.Optional[str] = None,
+        service_type: ty.Optional[str] = None,
+        interface: ty.Optional[str] = None,
+        region_name: ty.Optional[str] = None,
+        service_name: ty.Optional[str] = None,
+        version: ty.Optional[str] = None,
+        allow: ty.Optional[ty.Dict[str, ty.Any]] = None,
+        allow_version_hack: bool = True,
+        skip_discovery: bool = False,
+        discover_versions: bool = False,
+        min_version: ty.Optional[discover._RAW_VERSION_T] = None,
+        max_version: ty.Optional[discover._RAW_VERSION_T] = None,
+        **kwargs: ty.Any,
+    ) -> ty.Optional[discover._PARSED_VERSION_T]:
         """Return the major API version for a service.
 
         If a valid token is not present then a new one will be fetched using
@@ -399,6 +464,7 @@ class BaseIdentityPlugin(plugin.BaseAuthPlugin):
 
         :param session: A session object that can be used for communication.
         :type session: keystoneauth1.session.Session
+        :param str endpoint_override: URL to use for version discovery.
         :param string service_type: The type of service to lookup the endpoint
                                     for. This plugin will return None (failure)
                                     if service_type is not provided.
@@ -458,8 +524,8 @@ class BaseIdentityPlugin(plugin.BaseAuthPlugin):
         ``block-storage`` service and they do::
 
           client = adapter.Adapter(
-              session, service_type='block-storage', min_version=2,
-              max_version=3)
+              session, service_type='block-storage', min_version=2, max_version=3
+          )
           volume_version = client.get_api_major_version()
 
         The version actually be returned with no api calls other than getting
@@ -487,15 +553,24 @@ class BaseIdentityPlugin(plugin.BaseAuthPlugin):
         # Explode `version` into min_version and max_version - everything below
         # here uses the latter rather than the former.
         min_version, max_version = discover._normalize_version_args(
-            version, min_version, max_version)
+            version, min_version, max_version, service_type=service_type
+        )
         # Using functools.partial here just to reduce copy-pasta of params
         get_endpoint_data = functools.partial(
             self.get_endpoint_data,
-            session, service_type=service_type, interface=interface,
-            region_name=region_name, service_name=service_name,
-            allow=allow, min_version=min_version, max_version=max_version,
+            session,
+            endpoint_override=endpoint_override,
+            service_type=service_type,
+            interface=interface,
+            region_name=region_name,
+            service_name=service_name,
+            allow=allow,
+            min_version=min_version,
+            max_version=max_version,
             skip_discovery=skip_discovery,
-            allow_version_hack=allow_version_hack, **kwargs)
+            allow_version_hack=allow_version_hack,
+            **kwargs,
+        )
         data = get_endpoint_data(discover_versions=discover_versions)
         if (not data or not data.api_version) and not discover_versions:
             # It's possible that no version was requested and the endpoint
@@ -507,27 +582,109 @@ class BaseIdentityPlugin(plugin.BaseAuthPlugin):
             return None
         return data.api_version
 
-    def get_user_id(self, session, **kwargs):
+    def get_all_version_data(
+        self,
+        session: ks_session.Session,
+        interface: str = 'public',
+        region_name: ty.Optional[str] = None,
+        service_type: ty.Optional[str] = None,
+        **kwargs: ty.Any,
+    ) -> ty.Dict[
+        str, ty.Dict[str, ty.Dict[str, ty.List[discover.VersionData]]]
+    ]:
+        """Get version data for all services in the catalog.
+
+        :param session: A session object that can be used for communication.
+        :type session: keystoneauth1.session.Session
+        :param interface:
+            Type of endpoint to get version data for. Can be a single value
+            or a list of values. A value of None indicates that all interfaces
+            should be queried. (optional, defaults to public)
+        :param string region_name:
+            Region of endpoints to get version data for. A valueof None
+            indicates that all regions should be queried. (optional, defaults
+            to None)
+        :param string service_type:
+            Limit the version data to a single service. (optional, defaults
+            to None)
+        :returns: A dictionary keyed by region_name with values containing
+            dictionaries keyed by interface with values being a list of
+            :class:`~keystoneauth1.discover.VersionData`.
+        """
+        service_types = discover._SERVICE_TYPES
+        catalog = self.get_access(session).service_catalog
+        version_data: ty.Dict[
+            str, ty.Dict[str, ty.Dict[str, ty.List[discover.VersionData]]]
+        ] = {}
+        endpoints_data = catalog.get_endpoints_data(
+            interface=interface,
+            region_name=region_name,
+            service_type=service_type,
+        )
+
+        for endpoint_service_type, services in endpoints_data.items():
+            if service_types.is_known(endpoint_service_type):
+                endpoint_service_type = service_types.get_service_type(
+                    endpoint_service_type
+                )
+
+            for service in services:
+                if not service.region_name or not service.interface:
+                    continue
+
+                versions = service.get_all_version_string_data(
+                    session=session, project_id=self.get_project_id(session)
+                )
+
+                if service.region_name not in version_data:
+                    version_data[service.region_name] = {}
+                regions = version_data[service.region_name]
+
+                interface = service.interface.rstrip('URL')
+                if interface not in regions:
+                    regions[interface] = {}
+                regions[interface][endpoint_service_type] = versions
+
+        return version_data
+
+    def get_user_id(
+        self, session: ks_session.Session, **kwargs: ty.Any
+    ) -> ty.Optional[str]:
         return self.get_access(session).user_id
 
-    def get_project_id(self, session, **kwargs):
+    def get_project_id(
+        self, session: ks_session.Session, **kwargs: ty.Any
+    ) -> ty.Optional[str]:
         return self.get_access(session).project_id
 
-    def get_sp_auth_url(self, session, sp_id, **kwargs):
+    def get_sp_auth_url(
+        self, session: ks_session.Session, sp_id: str, **kwargs: ty.Any
+    ) -> ty.Optional[str]:
+        service_providers = self.get_access(session).service_providers
+        if not service_providers:
+            return None
         try:
-            return self.get_access(
-                session).service_providers.get_auth_url(sp_id)
+            return service_providers.get_auth_url(sp_id)
         except exceptions.ServiceProviderNotFound:
             return None
 
-    def get_sp_url(self, session, sp_id, **kwargs):
+    def get_sp_url(
+        self, session: ks_session.Session, sp_id: str, **kwargs: ty.Any
+    ) -> ty.Optional[str]:
+        service_providers = self.get_access(session).service_providers
+        if not service_providers:
+            return None
         try:
-            return self.get_access(
-                session).service_providers.get_sp_url(sp_id)
+            return service_providers.get_sp_url(sp_id)
         except exceptions.ServiceProviderNotFound:
             return None
 
-    def get_discovery(self, session, url, authenticated=None):
+    def get_discovery(
+        self,
+        session: ks_session.Session,
+        url: str,
+        authenticated: ty.Optional[bool] = None,
+    ) -> discover.Discover:
         """Return the discovery object for a URL.
 
         Check the session and the plugin cache to see if we have already
@@ -551,11 +708,14 @@ class BaseIdentityPlugin(plugin.BaseAuthPlugin):
 
         :returns: A discovery object with the results of looking up that URL.
         """
-        return discover.get_discovery(session=session, url=url,
-                                      cache=self._discovery_cache,
-                                      authenticated=authenticated)
+        return discover.get_discovery(
+            session=session,
+            url=url,
+            cache=self._discovery_cache,
+            authenticated=authenticated,
+        )
 
-    def get_cache_id_elements(self):
+    def get_cache_id_elements(self) -> ty.Dict[str, ty.Optional[str]]:
         """Get the elements for this auth plugin that make it unique.
 
         As part of the get_cache_id requirement we need to determine what
@@ -570,7 +730,7 @@ class BaseIdentityPlugin(plugin.BaseAuthPlugin):
         """
         raise NotImplementedError()
 
-    def get_cache_id(self):
+    def get_cache_id(self) -> ty.Optional[str]:
         """Fetch an identifier that uniquely identifies the auth options.
 
         The returned identifier need not be decomposable or otherwise provide
@@ -592,18 +752,12 @@ class BaseIdentityPlugin(plugin.BaseAuthPlugin):
 
         for k, v in sorted(elements.items()):
             if v is not None:
-                # NOTE(jamielennox): in python3 you need to pass bytes to hash
-                if isinstance(k, six.string_types):
-                    k = k.encode('utf-8')
-                if isinstance(v, six.string_types):
-                    v = v.encode('utf-8')
-
-                hasher.update(k)
-                hasher.update(v)
+                hasher.update(k.encode('utf-8'))
+                hasher.update(v.encode('utf-8'))
 
         return base64.b64encode(hasher.digest()).decode('utf-8')
 
-    def get_auth_state(self):
+    def get_auth_state(self) -> ty.Optional[str]:
         """Retrieve the current authentication state for the plugin.
 
         Retrieve any internal state that represents the authenticated plugin.
@@ -615,13 +769,17 @@ class BaseIdentityPlugin(plugin.BaseAuthPlugin):
                   set_auth_state to set the same authentication.
         :rtype: str or None if no auth present.
         """
-        if self.auth_ref:
-            data = {'auth_token': self.auth_ref.auth_token,
-                    'body': self.auth_ref._data}
+        if not self.auth_ref:
+            return None
 
-            return json.dumps(data)
+        data = {
+            'auth_token': self.auth_ref.auth_token,
+            'body': self.auth_ref._data,
+        }
 
-    def set_auth_state(self, data):
+        return json.dumps(data)
+
+    def set_auth_state(self, data: str) -> None:
         """Install existing authentication state for a plugin.
 
         Take the output of get_auth_state and install that authentication state
@@ -629,7 +787,8 @@ class BaseIdentityPlugin(plugin.BaseAuthPlugin):
         """
         if data:
             auth_data = json.loads(data)
-            self.auth_ref = access.create(body=auth_data['body'],
-                                          auth_token=auth_data['auth_token'])
+            self.auth_ref = access.create(
+                body=auth_data['body'], auth_token=auth_data['auth_token']
+            )
         else:
             self.auth_ref = None
