@@ -11,29 +11,38 @@
 # under the License.
 
 import abc
+import typing as ty
 
-import six
 import stevedore
+from stevedore import extension
 
 from keystoneauth1 import exceptions
+from keystoneauth1 import plugin
+
+if ty.TYPE_CHECKING:
+    from keystoneauth1.loading import opts
 
 PLUGIN_NAMESPACE = 'keystoneauth1.plugin'
 
 
-__all__ = ('get_available_plugin_names',
-           'get_available_plugin_loaders',
-           'get_plugin_loader',
-           'get_plugin_options',
-           'BaseLoader',
-           'PLUGIN_NAMESPACE')
+__all__ = (
+    'get_available_plugin_names',
+    'get_available_plugin_loaders',
+    'get_plugin_loader',
+    'get_plugin_options',
+    'BaseLoader',
+    'PLUGIN_NAMESPACE',
+)
+
+T = ty.TypeVar('T', covariant=True)
 
 
-def _auth_plugin_available(ext):
+def _auth_plugin_available(ext: extension.Extension) -> bool:
     """Read the value of available for whether to load this plugin."""
-    return ext.obj.available
+    return ty.cast(bool, ext.obj.available)
 
 
-def get_available_plugin_names():
+def get_available_plugin_names() -> ty.FrozenSet[str]:
     """Get the names of all the plugins that are available on the system.
 
     This is particularly useful for help and error text to prompt a user for
@@ -42,29 +51,35 @@ def get_available_plugin_names():
     :returns: A list of names.
     :rtype: frozenset
     """
-    mgr = stevedore.EnabledExtensionManager(namespace=PLUGIN_NAMESPACE,
-                                            check_func=_auth_plugin_available,
-                                            invoke_on_load=True,
-                                            propagate_map_exceptions=True)
+    mgr = stevedore.EnabledExtensionManager(
+        namespace=PLUGIN_NAMESPACE,
+        check_func=_auth_plugin_available,
+        invoke_on_load=True,
+        propagate_map_exceptions=True,
+    )
     return frozenset(mgr.names())
 
 
-def get_available_plugin_loaders():
+def get_available_plugin_loaders() -> (
+    ty.Dict[str, 'BaseLoader[plugin.BaseAuthPluginT]']
+):
     """Retrieve all the plugin classes available on the system.
 
     :returns: A dict with plugin entrypoint name as the key and the plugin
               loader as the value.
     :rtype: dict
     """
-    mgr = stevedore.EnabledExtensionManager(namespace=PLUGIN_NAMESPACE,
-                                            check_func=_auth_plugin_available,
-                                            invoke_on_load=True,
-                                            propagate_map_exceptions=True)
+    mgr = stevedore.EnabledExtensionManager(
+        namespace=PLUGIN_NAMESPACE,
+        check_func=_auth_plugin_available,
+        invoke_on_load=True,
+        propagate_map_exceptions=True,
+    )
 
     return dict(mgr.map(lambda ext: (ext.entry_point.name, ext.obj)))
 
 
-def get_plugin_loader(name):
+def get_plugin_loader(name: str) -> 'BaseLoader[plugin.BaseAuthPluginT]':
     """Retrieve a plugin class by its entrypoint name.
 
     :param str name: The name of the object to get.
@@ -76,16 +91,16 @@ def get_plugin_loader(name):
         if a plugin cannot be created.
     """
     try:
-        mgr = stevedore.DriverManager(namespace=PLUGIN_NAMESPACE,
-                                      invoke_on_load=True,
-                                      name=name)
+        mgr = stevedore.DriverManager(
+            namespace=PLUGIN_NAMESPACE, invoke_on_load=True, name=name
+        )
     except RuntimeError:
         raise exceptions.NoMatchingPlugin(name)
 
-    return mgr.driver
+    return ty.cast('BaseLoader[plugin.BaseAuthPluginT]', mgr.driver)
 
 
-def get_plugin_options(name):
+def get_plugin_options(name: str) -> ty.List['opts.Opt']:
     """Get the options for a specific plugin.
 
     This will be the list of options that is registered and loaded by the
@@ -99,14 +114,12 @@ def get_plugin_options(name):
     return get_plugin_loader(name).get_options()
 
 
-@six.add_metaclass(abc.ABCMeta)
-class BaseLoader(object):
-
+class _BaseLoader(ty.Generic[T], metaclass=abc.ABCMeta):
     @property
-    def plugin_class(self):
+    def plugin_class(self) -> ty.Type[T]:
         raise NotImplementedError()
 
-    def create_plugin(self, **kwargs):
+    def create_plugin(self, **kwargs: ty.Any) -> T:
         """Create a plugin from the options available for the loader.
 
         Given the options that were specified by the loader create an
@@ -125,7 +138,7 @@ class BaseLoader(object):
         return self.plugin_class(**kwargs)
 
     @abc.abstractmethod
-    def get_options(self):
+    def get_options(self) -> ty.List['opts.Opt']:
         """Return the list of parameters associated with the auth plugin.
 
         This list may be used to generate CLI or config arguments.
@@ -137,7 +150,7 @@ class BaseLoader(object):
         return []
 
     @property
-    def available(self):
+    def available(self) -> bool:
         """Return if the plugin is available for loading.
 
         If a plugin is missing dependencies or for some other reason should not
@@ -148,22 +161,27 @@ class BaseLoader(object):
         """
         return True
 
-    def load_from_options(self, **kwargs):
+    def load_from_options(self, **kwargs: ty.Any) -> T:
         """Create a plugin from the arguments retrieved from get_options.
 
         A client can override this function to do argument validation or to
         handle differences between the registered options and what is required
         to create the plugin.
         """
-        missing_required = [o for o in self.get_options()
-                            if o.required and kwargs.get(o.dest) is None]
+        missing_required = [
+            o
+            for o in self.get_options()
+            if o.required and kwargs.get(o.dest) is None
+        ]
 
         if missing_required:
             raise exceptions.MissingRequiredOptions(missing_required)
 
         return self.create_plugin(**kwargs)
 
-    def load_from_options_getter(self, getter, **kwargs):
+    def load_from_options_getter(
+        self, getter: ty.Callable[['opts.Opt'], ty.Any], **kwargs: ty.Any
+    ) -> T:
         """Load a plugin from getter function that returns appropriate values.
 
         To handle cases other than the provided CONF and CLI loading you can
@@ -185,3 +203,6 @@ class BaseLoader(object):
             kwargs[opt.dest] = val
 
         return self.load_from_options(**kwargs)
+
+
+class BaseLoader(_BaseLoader[plugin.BaseAuthPluginT]): ...
