@@ -12,27 +12,35 @@
 
 import argparse
 import os
-
-from positional import positional
+import typing as ty
 
 from keystoneauth1.loading import base
 
-
-__all__ = ('register_argparse_arguments',
-           'load_from_argparse_arguments')
+if ty.TYPE_CHECKING:
+    from keystoneauth1.loading import opts
+    from keystoneauth1 import plugin
 
 
-def _register_plugin_argparse_arguments(parser, plugin):
+__all__ = ('register_argparse_arguments', 'load_from_argparse_arguments')
+
+
+def _register_plugin_argparse_arguments(
+    parser: ty.Union[argparse.ArgumentParser, argparse._ArgumentGroup],
+    plugin: base.BaseLoader['plugin.BaseAuthPluginT'],
+) -> None:
     for opt in plugin.get_options():
-        parser.add_argument(*opt.argparse_args,
-                            default=opt.argparse_default,
-                            metavar=opt.metavar,
-                            help=opt.help,
-                            dest='os_%s' % opt.dest)
+        parser.add_argument(
+            *opt.argparse_args,
+            default=opt.argparse_default,
+            metavar=opt.metavar,
+            help=opt.help,
+            dest=f'os_{opt.dest}',
+        )
 
 
-@positional()
-def register_argparse_arguments(parser, argv, default=None):
+def register_argparse_arguments(
+    parser: argparse.ArgumentParser, argv: ty.List[str], default: ty.Any = None
+) -> ty.Optional[base.BaseLoader['plugin.BaseAuthPluginT']]:
     """Register CLI options needed to create a plugin.
 
     The function inspects the provided arguments so that it can also register
@@ -45,20 +53,23 @@ def register_argparse_arguments(parser, argv, default=None):
                               if one isn't specified by the CLI. default: None.
 
     :returns: The plugin class that will be loaded or None if not provided.
-    :rtype: :class:`keystoneauth1.plugin.BaseAuthPlugin`
+    :rtype: :class:`keystoneauth1.loader.BaseLoader`
 
     :raises keystoneauth1.exceptions.auth_plugins.NoMatchingPlugin:
         if a plugin cannot be created.
     """
     in_parser = argparse.ArgumentParser(add_help=False)
-    env_plugin = os.environ.get('OS_AUTH_TYPE',
-                                os.environ.get('OS_AUTH_PLUGIN', default))
+    env_plugin = os.environ.get(
+        'OS_AUTH_TYPE', os.environ.get('OS_AUTH_PLUGIN', default)
+    )
     for p in (in_parser, parser):
-        p.add_argument('--os-auth-type',
-                       '--os-auth-plugin',
-                       metavar='<name>',
-                       default=env_plugin,
-                       help='Authentication type to use')
+        p.add_argument(
+            '--os-auth-type',
+            '--os-auth-plugin',
+            metavar='<name>',
+            default=env_plugin,
+            help='Authentication type to use',
+        )
 
     options, _args = in_parser.parse_known_args(argv)
 
@@ -69,7 +80,7 @@ def register_argparse_arguments(parser, argv, default=None):
         msg = 'Default Authentication options'
         plugin = options.os_auth_type
     else:
-        msg = 'Options specific to the %s plugin.' % options.os_auth_type
+        msg = f'Options specific to the {options.os_auth_type} plugin.'
         plugin = base.get_plugin_loader(options.os_auth_type)
 
     group = parser.add_argument_group('Authentication Options', msg)
@@ -77,7 +88,9 @@ def register_argparse_arguments(parser, argv, default=None):
     return plugin
 
 
-def load_from_argparse_arguments(namespace, **kwargs):
+def load_from_argparse_arguments(
+    namespace: argparse.Namespace, **kwargs: ty.Any
+) -> ty.Optional['plugin.BaseAuthPluginT']:
     """Retrieve the created plugin from the completed argparse results.
 
     Loads and creates the auth plugin from the information parsed from the
@@ -91,15 +104,18 @@ def load_from_argparse_arguments(namespace, **kwargs):
     :raises keystoneauth1.exceptions.auth_plugins.NoMatchingPlugin:
         if a plugin cannot be created.
     """
-    if not namespace.os_auth_type:
+    os_auth_type = namespace.os_auth_type
+
+    if not os_auth_type:
         return None
 
-    if isinstance(namespace.os_auth_type, type):
-        plugin = namespace.os_auth_type
+    loader: base.BaseLoader[plugin.BaseAuthPluginT]
+    if isinstance(os_auth_type, base.BaseLoader):
+        loader = os_auth_type
     else:
-        plugin = base.get_plugin_loader(namespace.os_auth_type)
+        loader = base.get_plugin_loader(os_auth_type)
 
-    def _getter(opt):
-        return getattr(namespace, 'os_%s' % opt.dest)
+    def _getter(opt: 'opts.Opt') -> ty.Any:
+        return getattr(namespace, f'os_{opt.dest}')
 
-    return plugin.load_from_options_getter(_getter)
+    return loader.load_from_options_getter(_getter, **kwargs)
