@@ -10,28 +10,22 @@
 # License for the specific language governing permissions and limitations
 # under the License.
 
+import typing as ty
+
 from keystoneauth1 import plugin
 
+if ty.TYPE_CHECKING:
+    from keystoneauth1 import session as ks_session
 
-class NoAuth(plugin.BaseAuthPlugin):
+
+class NoAuth(plugin.FixedEndpointPlugin):
     """A provider that will always use no auth.
 
     This is useful to unify session/adapter loading for services
     that might be deployed in standalone/noauth mode.
     """
 
-    def __init__(self, endpoint=None):
-        super(NoAuth, self).__init__()
-        self.endpoint = endpoint
-
-    def get_token(self, session, **kwargs):
+    def get_token(
+        self, session: 'ks_session.Session', **kwargs: ty.Any
+    ) -> ty.Optional[str]:
         return 'notused'
-
-    def get_endpoint(self, session, **kwargs):
-        """Return the supplied endpoint.
-
-        Using this plugin the same endpoint is returned regardless of the
-        parameters passed to the plugin. endpoint_override overrides the
-        endpoint specified when constructing the plugin.
-        """
-        return kwargs.get('endpoint_override') or self.endpoint
